import os
import posixpath
from collections.abc import Iterable, Iterator
from itertools import zip_longest
from typing import TYPE_CHECKING, Optional

from fsspec.asyn import get_loop, sync
from sqlalchemy.sql import func
from tqdm import tqdm

from dvcx.node import DirType, Entry, Node, NodeWithPath
from dvcx.utils import TIME_ZERO, suffix_to_number

if TYPE_CHECKING:
    from dvcx.catalog.datasource import DataSource
    from dvcx.client import Client
    from dvcx.data_storage import AbstractMetastore, AbstractWarehouse
    from dvcx.dataset import DatasetRecord
    from dvcx.storage import Storage


RANDOM_BITS = 63  # size of the random integer field


class Listing:
    def __init__(
        self,
        storage: Optional["Storage"],
        metastore: "AbstractMetastore",
        warehouse: "AbstractWarehouse",
        client: "Client",
        dataset: Optional["DatasetRecord"],
    ):
        self.storage = storage
        self.metastore = metastore
        self.warehouse = warehouse
        self.client = client
        self.dataset = dataset  # dataset representing bucket listing

    def clone(self) -> "Listing":
        return self.__class__(
            self.storage,
            self.metastore.clone(),
            self.warehouse.clone(),
            self.client,
            self.dataset,
        )

    @property
    def id(self):
        return self.storage.id

    @property
    def dataset_rows(self):
        return self.warehouse.dataset_rows(self.dataset, self.dataset.latest_version)

    def fetch(self, start_prefix="") -> None:
        sync(get_loop(), self._fetch, start_prefix)

    async def _fetch(self, start_prefix) -> None:
        self = self.clone()
        if start_prefix:
            start_prefix = start_prefix.rstrip("/")
            base_entry = Entry.from_dir(
                posixpath.dirname(start_prefix),
                posixpath.basename(start_prefix),
                last_modified=TIME_ZERO,
            )
        else:
            base_entry = Entry.root()
        try:
            self.insert_entry(base_entry)
            async for entries in self.client.scandir(start_prefix):
                self.insert_entries(entries)
                if len(entries) > 1:
                    self.metastore.update_last_inserted_at()
        finally:
            self.insert_entries_done()

    def insert_entry(self, entry: Entry) -> None:
        self.warehouse.insert_rows(
            self.dataset_rows.get_table(),
            self.warehouse.prepare_entries(self.client.uri, [entry]),
        )

    def insert_entries(self, entries: Iterable[Entry]) -> None:
        self.warehouse.insert_rows(
            self.dataset_rows.get_table(),
            self.warehouse.prepare_entries(self.client.uri, entries),
        )

    def insert_entries_done(self) -> None:
        self.warehouse.insert_rows_done(self.dataset_rows.get_table())

    def expand_path(self, path) -> list[Node]:
        return self.warehouse.expand_path(self.dataset_rows, path)

    def resolve_path(self, path) -> Node:
        return self.warehouse.get_node_by_path(self.dataset_rows, path)

    def ls_path(self, node, fields):
        if node.vtype == "tar" or node.dir_type == DirType.TAR_ARCHIVE:
            return self.warehouse.select_node_fields_by_parent_path_tar(
                self.dataset_rows, node.path, fields
            )
        return self.warehouse.select_node_fields_by_parent_path(
            self.dataset_rows, node.path, fields
        )

    def collect_nodes_to_instantiate(
        self,
        sources: Iterable["DataSource"],
        copy_to_filename: Optional[str],
        recursive=False,
        copy_dir_contents=False,
        relative_path=None,
        from_edvcx=False,
        from_dataset=False,
    ) -> list[NodeWithPath]:
        rel_path_elements = relative_path.split("/") if relative_path else []
        all_nodes: list[NodeWithPath] = []
        for src in sources:
            node = src.node
            if recursive and src.is_container():
                dir_path = []
                if not copy_dir_contents:
                    dir_path.append(node.name)
                subtree_nodes = src.find(sort=["parent", "name"])
                all_nodes.extend(
                    NodeWithPath(n.n, path=dir_path + n.path) for n in subtree_nodes
                )
            else:
                node_path = []
                if from_edvcx:
                    for rpe, npe in zip_longest(
                        rel_path_elements, node.path.split("/")
                    ):
                        if rpe == npe:
                            continue
                        if npe:
                            node_path.append(npe)
                elif copy_to_filename:
                    node_path = [os.path.basename(copy_to_filename)]
                elif from_dataset:
                    node_path = [
                        src.listing.client.name,
                        node.parent,
                        node.name,
                    ]
                else:
                    node_path = [node.name]
                all_nodes.append(NodeWithPath(node, path=node_path))
        return all_nodes

    def instantiate_nodes(
        self,
        all_nodes,
        output,
        total_files=None,
        force=False,
        shared_progress_bar=None,
    ):
        progress_bar = shared_progress_bar or tqdm(
            desc=f"Instantiating '{output}'",
            unit=" files",
            unit_scale=True,
            unit_divisor=1000,
            total=total_files,
        )

        counter = 0
        for node in all_nodes:
            dst = os.path.join(output, *node.path)
            dst_dir = os.path.dirname(dst)
            os.makedirs(dst_dir, exist_ok=True)
            uid = node.n.as_uid(self.client.uri)
            self.client.instantiate_object(uid, dst, progress_bar, force)
            counter += 1
            if counter > 1000:
                progress_bar.update(counter)
                counter = 0

        progress_bar.update(counter)

    def find(
        self,
        node,
        fields,
        names=None,
        inames=None,
        paths=None,
        ipaths=None,
        size=None,
        type=None,
        order_by=None,
    ):
        dr = self.dataset_rows
        conds = []
        if names:
            for name in names:
                conds.append(dr.c.name.op("GLOB")(name))
        if inames:
            for iname in inames:
                conds.append(func.lower(dr.c.name).op("GLOB")(iname.lower()))
        if paths:
            node_path = self.warehouse.path_expr(dr)
            for path in paths:
                conds.append(node_path.op("GLOB")(path))
        if ipaths:
            node_path = self.warehouse.path_expr(dr)
            for ipath in ipaths:
                conds.append(func.lower(node_path).op("GLOB")(ipath.lower()))

        if size is not None:
            size_limit = suffix_to_number(size)
            if size_limit >= 0:
                conds.append(dr.c.size >= size_limit)
            else:
                conds.append(dr.c.size <= -size_limit)

        return self.warehouse.find(
            dr,
            node,
            fields,
            type=type,
            conds=conds,
            order_by=order_by,
        )

    def du(self, node: Node, count_files: bool = False):
        return self.warehouse.size(self.dataset_rows, node, count_files)

    def subtree_files(self, node: Node, sort=None):
        if node.dir_type == DirType.TAR_ARCHIVE or node.vtype != "":
            include_subobjects = True
        else:
            include_subobjects = False

        return self.warehouse.get_subtree_files(
            self.dataset_rows,
            node,
            sort=sort,
            include_subobjects=include_subobjects,
        )

    def get_nodes_by_parent_path(
        self,
        parent_path: str,
        type: Optional[str] = None,
    ) -> Iterator[Node]:
        return self.warehouse.get_nodes_by_parent_path(
            self.dataset_rows, parent_path, type="dir"
        )



import builtins
import json
from collections.abc import Iterable
from dataclasses import dataclass, fields
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

import attrs
from dateutil.parser import isoparse

from dvcx.client import Client
from dvcx.sql.types import NAME_TYPES_MAPPING, SQLType
from dvcx.utils import JSONSerialize

from .cache import UniqueId

if TYPE_CHECKING:
    from dvcx.storage import StorageURI

T = TypeVar("T", bound="DatasetRecord")
V = TypeVar("V", bound="DatasetVersion")
DD = TypeVar("DD", bound="DatasetDependency")

DATASET_PREFIX = "ds://"
QUERY_DATASET_PREFIX = "ds_query_"


def parse_dataset_uri(uri: str) -> tuple[str, Optional[int]]:
    """
    Parse dataser uri to extract name and version out of it (if version is defined)
    Example:
        Input: ds://zalando@v3
        Output: (zalando, 3)
    """
    p = urlparse(uri)
    if p.scheme != "ds":
        raise Exception("Dataset uri should start with ds://")
    s = p.netloc.split("@v")
    name = s[0]
    if len(s) == 1:
        return name, None
    if len(s) != 2:
        raise Exception(
            "Wrong dataset uri format, it should be: ds://<name>@v<version>"
        )
    version = int(s[1])
    return name, version


def create_dataset_uri(name: str, version: Optional[int] = None) -> str:
    """
    Creates a dataset uri based on dataset name and optionally version
    Example:
        Input: zalando, 3
        Output: ds//zalando@v3
    """
    uri = f"{DATASET_PREFIX}{name}"
    if version:
        uri += f"@v{version}"

    return uri


class DatasetDependencyType:
    DATASET = "dataset"
    STORAGE = "storage"


@dataclass
class DatasetDependency:
    id: int
    type: str
    name: str  # when the type is STORAGE, this is actually StorageURI
    version: str  # string until we'll have proper bucket listing versions
    created_at: datetime
    dependencies: list[Optional["DatasetDependency"]]

    @classmethod
    def parse(
        cls: builtins.type[DD],
        id: int,
        dataset_id: Optional[int],
        dataset_version_id: Optional[int],
        bucket_id: Optional[int],
        bucket_version: Optional[str],
        dataset_name: Optional[str],
        dataset_created_at: Optional[datetime],
        dataset_version: Optional[int],
        dataset_version_created_at: Optional[datetime],
        bucket_uri: Optional["StorageURI"],
    ) -> Optional["DatasetDependency"]:
        if dataset_id:
            assert dataset_name is not None
            return cls(
                id,
                DatasetDependencyType.DATASET,
                dataset_name,
                (
                    str(dataset_version)  # type: ignore[arg-type]
                    if dataset_version
                    else None
                ),
                dataset_version_created_at or dataset_created_at,  # type: ignore[arg-type]
                [],
            )
        if bucket_uri:
            return cls(
                id,
                DatasetDependencyType.STORAGE,
                bucket_uri,
                bucket_version,  # type: ignore[arg-type]
                isoparse(bucket_version),  # type: ignore[arg-type]
                [],
            )
        # dependency has been removed
        # TODO we should introduce flags for removed datasets, instead of
        # removing them from tables so that we can still have references
        return None

    @property
    def is_dataset(self) -> bool:
        return self.type == DatasetDependencyType.DATASET

    def __eq__(self, other):
        if not isinstance(other, DatasetDependency):
            return False

        return (
            self.type == other.type
            and self.name == other.name
            and self.version == other.version
        )

    def __hash__(self):
        return hash(f"{self.type}_{self.name}_{self.version}")


@dataclass
class DatasetStats:
    num_objects: Optional[int]  # None if table is missing
    size: Optional[int]  # in bytes None if table is missing or empty


class Status:
    CREATED = 1
    PENDING = 2
    FAILED = 3
    COMPLETE = 4
    STALE = 6


@dataclass
class DatasetVersion:
    id: int
    dataset_id: int
    version: int
    status: int
    created_at: datetime
    finished_at: Optional[datetime]
    error_message: str
    error_stack: str
    script_output: str
    custom_column_types: dict[str, Union[SQLType, type[SQLType]]]
    num_objects: Optional[int]
    size: Optional[int]
    preview: Optional[list[dict]]
    sources: str = ""
    query_script: str = ""

    @classmethod
    def parse(
        cls: type[V],
        id: int,
        dataset_id: int,
        version: int,
        status: int,
        created_at: datetime,
        finished_at: Optional[datetime],
        error_message: str,
        error_stack: str,
        script_output: str,
        num_objects: Optional[int],
        size: Optional[int],
        preview: Optional[str],
        custom_column_types: dict[str, Union[SQLType, type[SQLType]]],
        sources: str = "",
        query_script: str = "",
    ):
        return cls(
            id,
            dataset_id,
            version,
            status,
            created_at,
            finished_at,
            error_message,
            error_stack,
            script_output,
            custom_column_types,
            num_objects,
            size,
            json.loads(preview) if preview else None,
            sources,
            query_script,
        )

    def __eq__(self, other):
        if not isinstance(other, DatasetVersion):
            return False
        return self.version == other.version and self.dataset_id == other.dataset_id

    def __lt__(self, other):
        if not isinstance(other, DatasetVersion):
            return False
        return self.version < other.version

    def __hash__(self):
        return hash(f"{self.dataset_id}_{self.version}")

    def is_final_status(self) -> bool:
        return self.status in [Status.FAILED, Status.COMPLETE, Status.STALE]

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    @property
    def custom_column_types_serialized(self) -> dict[str, Any]:
        if not self.custom_column_types:
            return {}

        return {
            c_name: c_type.to_dict()
            if isinstance(c_type, SQLType)
            else c_type().to_dict()
            for c_name, c_type in self.custom_column_types.items()
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "DatasetVersion":
        kwargs = {f.name: d[f.name] for f in fields(cls) if f.name in d}
        return cls(**kwargs)


@dataclass
class DatasetRecord:
    id: int
    name: str
    description: Optional[str]
    labels: list[str]
    shadow: bool
    custom_column_types: dict[str, Union[SQLType, type[SQLType]]]
    versions: list[DatasetVersion]
    status: int = Status.CREATED
    created_at: Optional[datetime] = None
    finished_at: Optional[datetime] = None
    error_message: str = ""
    error_stack: str = ""
    script_output: str = ""
    sources: str = ""
    query_script: str = ""

    @staticmethod
    def parse_custom_column_types(
        ct: dict[str, Any],
    ) -> dict[str, Union[SQLType, type[SQLType]]]:
        return {
            c_name: NAME_TYPES_MAPPING[c_type["type"]].from_dict(c_type)  # type: ignore [attr-defined]
            for c_name, c_type in ct.items()
        }

    @classmethod
    def parse(  # noqa: PLR0913
        cls: type[T],
        id: int,
        name: str,
        description: Optional[str],
        labels: str,
        shadow: int,
        status: int,
        created_at: datetime,
        finished_at: Optional[datetime],
        error_message: str,
        error_stack: str,
        script_output: str,
        sources: str,
        query_script: str,
        custom_column_types: str,
        version_id: int,
        version_dataset_id: int,
        version: int,
        version_status: int,
        version_created_at: datetime,
        version_finished_at: Optional[datetime],
        version_error_message: str,
        version_error_stack: str,
        version_script_output: str,
        version_num_objects: Optional[int],
        version_size: Optional[int],
        version_preview: Optional[str],
        version_sources: Optional[str],
        version_query_script: Optional[str],
        version_custom_column_types: str,
    ) -> "DatasetRecord":
        labels_lst: list[str] = json.loads(labels) if labels else []
        custom_column_types_dct: dict[str, Any] = (
            json.loads(custom_column_types) if custom_column_types else {}
        )
        version_custom_column_types_dct: dict[str, str] = (
            json.loads(version_custom_column_types)
            if version_custom_column_types
            else {}
        )

        dataset_version = DatasetVersion.parse(
            version_id,
            version_dataset_id,
            version,
            version_status,
            version_created_at,
            version_finished_at,
            version_error_message,
            version_error_stack,
            version_script_output,
            version_num_objects,
            version_size,
            version_preview,
            cls.parse_custom_column_types(version_custom_column_types_dct),  # type: ignore[arg-type]
            version_sources,  # type: ignore[arg-type]
            version_query_script,  # type: ignore[arg-type]
        )

        return cls(
            id,
            name,
            description,
            labels_lst,
            bool(shadow),
            cls.parse_custom_column_types(custom_column_types_dct),  # type: ignore[arg-type]
            [dataset_version],
            status,
            created_at,
            finished_at,
            error_message,
            error_stack,
            script_output,
            sources,
            query_script,
        )

    @property
    def custom_column_types_serialized(self) -> dict[str, Any]:
        if not self.custom_column_types:
            return {}

        return {
            c_name: c_type.to_dict()
            if isinstance(c_type, SQLType)
            else c_type().to_dict()
            for c_name, c_type in self.custom_column_types.items()
        }

    def get_custom_column_types(
        self, version: int
    ) -> dict[str, Union[SQLType, type[SQLType]]]:
        return (
            self.get_version(version).custom_column_types
            if version
            else self.custom_column_types
        )

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def merge_versions(self, other: "DatasetRecord") -> "DatasetRecord":
        """Merge versions from another dataset"""
        if other.id != self.id:
            raise RuntimeError("Cannot merge versions of datasets with different ids")
        if not other.versions:
            # nothing to merge
            return self
        if not self.versions:
            self.versions = []

        self.versions = list(set(self.versions + other.versions))
        self.versions.sort(key=lambda v: v.version)
        return self

    def has_version(self, version: int) -> bool:
        return version in self.versions_values

    def is_valid_next_version(self, version: int) -> bool:
        """
        Checks if a number can be a valid next latest version for dataset.
        The only rule is that it cannot be lower than current latest version
        """
        if self.latest_version and self.latest_version >= version:
            return False
        return True

    def get_version(self, version: int) -> DatasetVersion:
        if not self.has_version(version):
            raise ValueError(f"Dataset {self.name} does not have version {version}")
        return next(
            v
            for v in self.versions  # type: ignore [union-attr]
            if v.version == version
        )

    def remove_version(self, version: int) -> None:
        if not self.versions or not self.has_version(version):
            return

        self.versions = [v for v in self.versions if v.version != version]

    def identifier(self, version: int) -> str:
        """
        Get identifier in the form my-dataset@v3
        """
        if not self.has_version(version):
            raise ValueError(f"Dataset {self.name} doesn't have a version {version}")
        return f"{self.name}@v{version}"

    def uri(self, version: int) -> str:
        """
        Dataset uri example: ds://dogs@v3
        """
        identifier = self.identifier(version)
        return f"{DATASET_PREFIX}{identifier}"

    @property
    def is_bucket_listing(self) -> bool:
        """
        For bucket listing we implicitly create underlying dataset to hold data. This
        method is checking if this is one of those datasets.
        """
        return Client.is_data_source_uri(self.name)

    @property
    def versions_values(self) -> list[int]:
        """
        Extracts actual versions from list of DatasetVersion objects
        in self.versions attribute
        """
        if not self.versions:
            return []

        return sorted(v.version for v in self.versions)

    @property
    def next_version(self) -> int:
        """Returns what should be next autoincrement version of dataset"""
        if not self.versions:
            return 1
        return max(self.versions_values) + 1

    @property
    def latest_version(self) -> int:
        """Returns latest version of a dataset"""
        return max(self.versions_values)

    @property
    def prev_version(self) -> Optional[int]:
        """Returns previous version of a dataset"""
        if len(self.versions) == 1:
            return None

        return sorted(self.versions_values)[-2]

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "DatasetRecord":
        versions = [DatasetVersion.from_dict(v) for v in d.pop("versions", [])]
        kwargs = {f.name: d[f.name] for f in fields(cls) if f.name in d}
        return cls(**kwargs, versions=versions)


@attrs.define
class DatasetRow:
    id: int
    vtype: str
    dir_type: int
    parent: str
    name: str
    etag: str
    version: str
    is_latest: bool
    last_modified: Optional[datetime]
    size: int
    owner_name: str
    owner_id: str
    random: int
    location: Optional[str]
    source: "StorageURI"
    custom: Optional[dict] = attrs.field(factory=dict)

    @classmethod
    def from_result_row(cls, columns: list[str], values: Iterable[Any]) -> "DatasetRow":
        row = dict(zip(columns, values))
        return cls.from_dict(row)

    @classmethod
    def from_dict(cls, row: dict[str, Any]) -> "DatasetRow":
        core_fields = {key: value for (key, value) in row.items() if hasattr(cls, key)}
        if core_fields.get("last_modified") and isinstance(
            core_fields["last_modified"], str
        ):
            core_fields["last_modified"] = isoparse(core_fields["last_modified"])
        custom_fields = {
            key: value for (key, value) in row.items() if key not in core_fields
        }
        core_fields["custom"] = custom_fields
        return cls(**core_fields)

    def __getitem__(self, col):
        if hasattr(self, col):
            return getattr(self, col)
        if self.custom and col in self.custom:
            return self.custom[col]
        raise KeyError

    def as_uid(self) -> UniqueId:
        return UniqueId(
            self.source,
            self.parent,
            self.name,
            self.etag,
            self.size,
            self.vtype,
            self.location,
        )

    def to_preview(self):
        return {
            "id": self.id,
            "vtype": self.vtype,
            "dir_type": self.dir_type,
            "parent": self.parent,
            "name": self.name,
            "etag": self.etag,
            "version": self.version,
            "is_latest": self.is_latest,
            "last_modified": (
                self.last_modified.isoformat() if self.last_modified else None
            ),
            "size": self.size,
            "owner_name": self.owner_name,
            "owner_id": self.owner_id,
            "random": self.random,
            "location": self.location,
            "source": self.source,
            "custom": (
                json.dumps(self.custom, cls=JSONSerialize) if self.custom else None
            ),
        }



import contextlib
import math
from abc import ABC, abstractmethod
from collections.abc import Generator, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Optional, Union

import sqlalchemy as sa

from dvcx.data_storage.schema import PARTITION_COLUMN_ID
from dvcx.data_storage.warehouse import SELECT_BATCH_SIZE

if TYPE_CHECKING:
    from dvcx.dataset import DatasetRow


@dataclass
class DatasetRowsBatch:
    rows: Sequence["DatasetRow"]


BatchingResult = Union["DatasetRow", DatasetRowsBatch]


class BatchingStrategy(ABC):
    """BatchingStrategy provides means of batching UDF executions."""

    @abstractmethod
    def __call__(
        self,
        execute: Callable,
        query: sa.sql.selectable.Select,
    ) -> Generator[BatchingResult, None, None]:
        """Apply the provided parameters to the UDF."""


class NoBatching(BatchingStrategy):
    """
    NoBatching implements the default batching strategy, which is not to
    batch UDF calls.
    """

    def __call__(
        self,
        execute: Callable,
        query: sa.sql.selectable.Select,
    ) -> Generator["DatasetRow", None, None]:
        return execute(query, limit=query._limit)


class Batch(BatchingStrategy):
    """
    Batch implements UDF call batching, where each execution of a UDF
    is passed a sequence of multiple parameter sets.
    """

    def __init__(self, count: int):
        self.count = count

    def __call__(
        self,
        execute: Callable,
        query: sa.sql.selectable.Select,
    ) -> Generator[DatasetRowsBatch, None, None]:
        # choose page size that is a multiple of the batch size
        page_size = math.ceil(SELECT_BATCH_SIZE / self.count) * self.count

        # select rows in batches
        results: list[DatasetRow] = []

        with contextlib.closing(
            execute(query, page_size=page_size, limit=query._limit)
        ) as rows:
            for row in rows:
                results.append(row)
                if len(results) >= self.count:
                    batch, results = results[: self.count], results[self.count :]
                    yield DatasetRowsBatch(batch)

            if len(results) > 0:
                yield DatasetRowsBatch(results)


class Partition(BatchingStrategy):
    """
    Partition implements UDF call batching, where each execution of a UDF
    is run on a list of dataset rows grouped by the specified column.
    Dataset rows need to be sorted by the grouping column.
    """

    def __call__(
        self,
        execute: Callable,
        query: sa.sql.selectable.Select,
    ) -> Generator[DatasetRowsBatch, None, None]:
        current_partition: Optional[int] = None
        batch: list[DatasetRow] = []

        with contextlib.closing(
            execute(query, order_by=(PARTITION_COLUMN_ID, "id"), limit=query._limit)
        ) as rows:
            for row in rows:
                partition = row[PARTITION_COLUMN_ID]
                if current_partition != partition:
                    current_partition = partition
                    if len(batch) > 0:
                        yield DatasetRowsBatch(batch)
                        batch = []
                batch.append(row)

            if len(batch) > 0:
                yield DatasetRowsBatch(batch)



import pytest

from dvcx.cache import DVCXCache, UniqueId


@pytest.fixture
def cache(tmp_path):
    return DVCXCache(str(tmp_path / "cache"), str(tmp_path / "tmp"))


def test_simple(cache):
    uid = UniqueId(
        "s3://foo", "data", "bar", etag="xyz", size=3, vtype="", location=None
    )
    data = b"foo"
    assert not cache.contains(uid)

    cache.store_data(uid, data)
    assert cache.contains(uid)
    with open(cache.get_path(uid), mode="rb") as f:
        assert f.read() == data

    cache.clear()
    assert not cache.contains(uid)


def test_get_total_size(cache):
    file_info = [
        ("file1", b"foo"),
        ("file2", b"bar"),
        ("file3", b"some file data"),
        ("file4", b"more file data " * 1024),
    ]
    expected_total = sum(len(d) for _, d in file_info)
    for name, data in file_info:
        uid = UniqueId(
            "s3://foo", "data", name, etag="xyz", size=3, vtype="", location=None
        )
        cache.store_data(uid, data)
    total = cache.get_total_size()
    assert total == expected_total

    cache.clear()
    empty_total = cache.get_total_size()
    assert empty_total == 0


def test_remove(cache):
    uid = UniqueId(
        "s3://bkt42", "dir1/dir2", "file84", etag="abc", size=3, vtype="", location=None
    )
    cache.store_data(uid, b"some random string 679")

    assert cache.contains(uid)
    cache.remove(uid)
    assert not cache.contains(uid)



import io
import json

import pytest

from dvcx.cache import UniqueId
from dvcx.catalog import Catalog
from dvcx.lib.file import File, FileInfo, TextFile
from dvcx.lib.utils import DvcxError


def test_uid_missing_location():
    name = "my_name"
    vtype = "vt1"

    stream = File(name=name, vtype=vtype)
    assert stream.get_uid() == UniqueId("", "", name, "", 0, vtype, None)


def test_uid_location():
    name = "na_me"
    vtype = "some_random"
    loc = {"e": 42}

    stream = File(name=name, vtype=vtype, location=loc)
    assert stream.get_uid() == UniqueId("", "", name, "", 0, vtype, loc)


def test_file_stem():
    s = File(name=".file.jpg.txt")
    assert s.get_file_stem() == ".file.jpg"


def test_file_ext():
    s = File(name=".file.jpg.txt")
    assert s.get_file_ext() == "txt"


def test_file_suffix():
    s = File(name=".file.jpg.txt")
    assert s.get_file_suffix() == ".txt"


def test_full_name():
    name = ".file.jpg.txt"
    f = File(name=name)
    assert f.get_full_name() == name

    parent = "dir1/dir2"
    f = File(name=name, parent=parent)
    assert f.get_full_name() == f"{parent}/{name}"


def test_cache_get_path(catalog: Catalog):
    stream = File(name="test.txt1", source="s3://mybkt")
    stream.set_catalog(catalog)

    uid = stream.get_uid()
    data = b"some data is heRe"
    catalog.cache.store_data(uid, data)

    path = stream.get_local_path()
    assert path is not None

    with open(path, mode="rb") as f:
        assert f.read() == data


def test_read_binary_data(tmp_path, catalog: Catalog):
    file_name = "myfile"
    data = b"some\x00data\x00is\x48\x65\x6c\x57\x6f\x72\x6c\x64\xff\xffheRe"

    file_path = tmp_path / file_name
    with open(file_path, "wb") as fd:
        fd.write(data)

    file = File(name=file_name, source=f"file://{tmp_path}")
    file.set_catalog(catalog)

    with open(file_path, "rb") as fd:
        file.set_file(fd, caching_enabled=False)
        assert file.read() == data


def test_read_binary_data_as_text(tmp_path, catalog: Catalog):
    file_name = "myfile43.txt"
    data = b"some\x00data\x00is\x48\x65\x6c\x57\x6f\x72\x6c\x64\xff\xffheRe"

    file_path = tmp_path / file_name
    with open(file_path, "wb") as fd:
        fd.write(data)

    file = TextFile(name=file_name, source=f"file://{tmp_path}")
    file.set_catalog(catalog)

    with open(file_path) as fd:
        file.set_file(fd, caching_enabled=False)
        try:
            x = file.read()
        except UnicodeDecodeError:  # Unix
            pass
        else:  # Windows
            assert x != data


def test_read_text_data(tmp_path, catalog: Catalog):
    file_name = "myfile"
    data = "this is a TexT data..."

    file_path = tmp_path / file_name
    with open(file_path, "w") as fd:
        fd.write(data)

    file = TextFile(name=file_name, source=f"file://{tmp_path}")
    file.set_catalog(catalog)

    with open(file_path) as fd:
        file.set_file(fd, caching_enabled=False)
        assert file.read() == data


def test_set_file_no_catalog(catalog):
    stream = File(name="name33")
    with pytest.raises(DvcxError):
        stream.set_file(io.StringIO(), False)


def test_cache_get_path_without_cache():
    stream = File(name="test.txt1", source="s3://mybkt")
    with pytest.raises(RuntimeError):
        stream.get_local_path()


def test_json_from_string():
    d = {"e": 12}

    file = File(name="something", location=d)
    assert file.location == d

    file = File(name="something", location=None)
    assert file.location is None

    file = File(name="something", location="")
    assert file.location is None

    file = File(name="something", location=json.dumps(d))
    assert file.location == d

    with pytest.raises(ValueError):
        File(name="something", location="{not a json}")


def test_file_info_jsons():
    file = FileInfo(name="something", location="")
    assert file.location is None

    d = {"e": 12}
    file = FileInfo(name="something", location=json.dumps(d))
    assert file.location == d



import json
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from random import getrandbits
from typing import ClassVar, Optional, Union

from pydantic import Field, field_validator

from dvcx.cache import UniqueId
from dvcx.client.fileslice import FileSlice
from dvcx.lib.cached_stream import PreCachedStream, PreDownloadStream
from dvcx.lib.feature import ShallowFeature
from dvcx.lib.utils import DvcxError
from dvcx.utils import TIME_ZERO


class FileFeature(ShallowFeature):
    _is_file = True

    def set_file(self, stream, caching_enabled: bool) -> None:
        raise NotImplementedError

    def open(self):
        raise NotImplementedError

    def read(self):
        with self.open() as stream:
            return stream.read()

    def get_value(self):
        return self.read()


class VFileError(DvcxError):
    def __init__(self, file: "File", message: str, vtype: str = ""):
        type_ = f" of vtype '{vtype}'" if vtype else ""
        super().__init__(f"Error in v-file '{file.get_uid().path}'{type_}: {message}")


class VFile(ABC):
    @classmethod
    @abstractmethod
    def get_vtype(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def open(cls, file: "File", location: list[dict]):
        pass


class TarVFile(VFile):
    @classmethod
    def get_vtype(cls) -> str:
        return "tar"

    @classmethod
    def open(cls, file: "File", location: list[dict]):
        if len(location) > 1:
            VFileError(file, "multiple 'location's are not supported yet")

        loc = location[0]

        if (offset := loc.get("offset", None)) is None:
            VFileError(file, "'offset' is not specified")

        if (size := loc.get("size", None)) is None:
            VFileError(file, "'size' is not specified")

        if (parent := loc.get("parent", None)) is None:
            VFileError(file, "'parent' is not specified")

        tar_file = File(**parent)
        tar_file.set_catalog(file._catalog)

        tar_file_uid = tar_file.get_uid()
        client = file._catalog.get_client(tar_file_uid.storage)
        fd = client.open_object(tar_file_uid, use_cache=file._caching_enabled)
        return FileSlice(fd, offset, size, file.name)


class VFileRegistry:
    _vtype_readers: ClassVar[dict[str, type["VFile"]]] = {"tar": TarVFile}

    @classmethod
    def register(cls, reader: type["VFile"]):
        cls._vtype_readers[reader.get_vtype()] = reader

    @classmethod
    def resolve(cls, file: "File", location: list[dict]):
        if len(location) == 0:
            raise VFileError(file, "'location' must not be list of JSONs")

        if not (vtype := location[0].get("vtype", "")):
            raise VFileError(file, "vtype is not specified")

        reader = cls._vtype_readers.get(vtype, None)
        if not reader:
            raise VFileError(file, "reader not registered", vtype)

        return reader.open(file, location)


class File(FileFeature):
    source: str = Field(default="")
    parent: str = Field(default="")
    name: str
    version: str = Field(default="")
    etag: str = Field(default="")
    size: int = Field(default=0)
    vtype: str = Field(default="")
    location: Optional[Union[dict, list[dict]]] = Field(default=None)

    _unique_id_keys: ClassVar[list[str]] = [
        "source",
        "parent",
        "name",
        "etag",
        "size",
        "vtype",
        "location",
    ]

    @staticmethod
    def to_dict(
        v: Optional[Union[str, dict, list[dict]]],
    ) -> Optional[Union[str, dict, list[dict]]]:
        if v is None or v == "":
            return None
        if isinstance(v, str):
            try:
                return json.loads(v)
            except Exception as e:  # noqa: BLE001
                raise ValueError(
                    f"Unable to convert string '{v}' to dict for File feature: {e}"
                ) from None
        return v

    # Workaround for empty JSONs converted to empty strings in some DBs.
    @field_validator("location", mode="before")
    @classmethod
    def validate_location(cls, v):
        return File.to_dict(v)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._stream = None
        self._catalog = None
        self._caching_enabled = False

    def open(self):
        if self.location:
            return VFileRegistry.resolve(self, self.location)

        return self._stream

    def set_catalog(self, catalog):
        self._catalog = catalog

    def set_file(self, stream, caching_enabled: bool) -> None:
        if self._catalog is None:
            raise DvcxError(f"Cannot set file '{stream}' without catalog")

        stream_class = PreCachedStream if caching_enabled else PreDownloadStream
        self._stream = stream_class(stream, self.size, self._catalog, self.get_uid())
        self._caching_enabled = caching_enabled

    def get_uid(self) -> UniqueId:
        dump = self.model_dump()
        return UniqueId(*(dump[k] for k in self._unique_id_keys))

    def get_local_path(self) -> Optional[str]:
        """Get path to a file in a local cache.
        Return None if file is not cached. Throws an exception if cache is not setup."""
        if self._catalog is None:
            raise RuntimeError(
                "cannot resolve local file path because catalog is not setup"
            )
        return self._catalog.cache.get_path(self.get_uid())

    def get_file_suffix(self):
        return Path(self.name).suffix

    def get_file_ext(self):
        return Path(self.name).suffix.strip(".")

    def get_file_stem(self):
        return Path(self.name).stem

    def get_full_name(self):
        if not self.parent:
            return self.name
        return f"{self.parent}/{self.name}"

    def get_full_path(self):
        return f"{self.source}/{self.get_full_name()}"


BinaryFile = File


class TextFile(File):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._file = None

    def set_file(self, stream, caching_enabled: bool) -> None:
        super().set_file(stream, caching_enabled)
        self._stream.set_mode("r")


class FileInfo(FileFeature):
    source: str = Field(default="")
    parent: str = Field(default="")
    name: str
    size: int = Field(default=0)
    location: Optional[Union[dict, list[dict]]] = Field(default=None)
    vtype: str = Field(default="")
    dir_type: int = Field(default=0)
    owner_name: str = Field(default="")
    owner_id: str = Field(default="")
    is_latest: bool = Field(default=True)
    last_modified: datetime = Field(default=TIME_ZERO)
    version: str = Field(default="")
    etag: str = Field(default="")
    random: int = Field(default_factory=lambda: getrandbits(63))

    @field_validator("location", mode="before")
    @classmethod
    def validate_location(cls, v):
        return File.to_dict(v)



from dvcx.sql.types import (
    TypeConverter,
    TypeDefaults,
    TypeReadConverter,
    register_backend_types,
    register_type_defaults,
    register_type_read_converters,
)

setup_is_complete: bool = False


def setup() -> None:
    global setup_is_complete  # noqa: PLW0603
    if setup_is_complete:
        return

    register_backend_types("default", TypeConverter())
    register_type_read_converters("default", TypeReadConverter())
    register_type_defaults("default", TypeDefaults())

    setup_is_complete = True



"""
To install dependencies:

  pip install imgbeddings

"""

import uuid

from imgbeddings import imgbeddings
from sqlalchemy import tuple_

from dvcx.lib.param import Image
from dvcx.query import C, DatasetQuery, udf
from dvcx.sql.functions.array import cosine_distance, euclidean_distance
from dvcx.sql.types import Array, Float32


@udf(
    params=(Image(),),
    output={"embedding": Array(Float32)},
    method="embedding",
)
class ImageEmbeddings:
    def __init__(self):
        self.emb = imgbeddings()

    def embedding(self, img):
        emb = self.emb.to_embeddings(img)
        return (emb[0].tolist(),)


ds1_name = uuid.uuid4().hex
print(f"Saving embeddings to dataset: {ds1_name}")
# Save as a new dataset
(
    DatasetQuery(path="gcs://dvcx-datalakes/dogs-and-cats/")
    .filter(C.name.glob("*cat*.jpg"))
    .add_signals(ImageEmbeddings)
    .save(ds1_name)
)

ds2_name = uuid.uuid4().hex
source, parent, name, embedding = (
    DatasetQuery(name=ds1_name)
    .select(C.source, C.parent, C.name, C.embedding)
    .order_by(C.source, C.parent, C.name)
    .limit(1)
    .results()[0]
)
(
    DatasetQuery(name=ds1_name)
    .filter(tuple_(C.source, C.parent, C.name) != (source, parent, name))
    .mutate(
        cos_dist=cosine_distance(C.embedding, embedding),
        eucl_dist=euclidean_distance(C.embedding, embedding),
    )
    .order_by(C.cos_dist)
    .limit(10)
    .select_except(C.embedding)
    .save(ds2_name)
)

print("target:", source, parent, name, embedding[:3])
print()
print("Top 10 by cosine distance:")
for row in (
    DatasetQuery(name=ds2_name)
    .select(C.source, C.parent, C.name, C.cos_dist, C.eucl_dist)
    .order_by(C.cos_dist)
    .results()
):
    print(*row)



import os

import pandas as pd

from dvcx.lib.claude import Claude, ClaudeMessage
from dvcx.lib.dataset import Dataset
from dvcx.lib.file import File
from dvcx.query.schema import C

SOURCE = "s3://ldb-public/remote/chatbots-public/"
MODEL = "claude-3-opus-20240229"
PROMPT = """Summarise the dialog in a sentence"""

ds = (
    Dataset(SOURCE, anon=True)
    .filter(C.name.glob("*.txt"))
    .limit(5)
    .map(
        ClaudeMessage(
            prompt=PROMPT,
            model_name=MODEL,
            temperature=0.9,
            api_key=os.getenv("ANTHROPIC_API_KEY"),
        ),
        parallel=3,
        cache=True,
    )
)

df = ds.select(File.name, Claude).to_pandas()

with pd.option_context("display.max_columns", None):
    print(df)



import os

import fsspec
from PIL import Image

from dvcx.catalog import get_catalog
from dvcx.query import DatasetRow, Object, udf


def load_image(raw):
    img = Image.open(raw)
    img.load()
    return img


@udf(
    output=DatasetRow.schema,
    params=(Object(load_image), *tuple(DatasetRow.schema.keys())),
)
class ImageTransform:
    def __init__(
        self,
        *,
        image_filter,
        bucket_name,
        prefix,
        output_folder,
        file_prefix="",
        vtype="",
    ):
        # Once we fix the UDF decorator situation, it would make more sense to put this
        # into a child class and make apply_filter an abstractmethod.
        self.image_filter = image_filter
        self.folder_name = output_folder
        self.file_prefix = file_prefix
        self.prefix = prefix
        self.vtype = vtype

        catalog = get_catalog()
        self.client, _ = catalog.parse_url(os.path.join(self.prefix, bucket_name))

    def apply_filter(self, image):
        return image.filter(self.image_filter)

    def save(self, image, source, name, format):
        # Make name for new image
        new_name = f"{self.file_prefix}{name}"

        # Do writeback
        blob_name = os.path.join(self.folder_name, new_name)
        urlpath = os.path.join(source, blob_name)
        cloud_file = fsspec.open(urlpath=urlpath, mode="wb")
        with cloud_file as fp:
            image.save(fp, format=format)

        # Get the blob info
        info_ = self.client.fs.info(urlpath)
        info = self.client.convert_info(info_, self.folder_name)
        info.name = new_name
        return info

    def __call__(
        self,
        image,
        *args,
    ):
        # Build a dict from row contents
        record = dict(zip(DatasetRow.schema.keys(), args))
        del record["random"]  # random will be populated automatically
        record["is_latest"] = record["is_latest"] > 0  # needs to be a bool

        # yield same row back
        yield DatasetRow.create(**record)

        # Don't apply the filter twice
        if record["parent"] == self.folder_name:
            return

        # Apply the filter
        image_b = self.apply_filter(image)

        # Save the image and get the cloud object info
        entry = self.save(
            image_b, record["source"], name=record["name"], format=image.format
        )

        # Build the new row
        yield DatasetRow.create(
            name=entry.name,
            source=record["source"],
            parent=self.folder_name,
            size=entry.size,
            location=record["name"]
            if not record["parent"]
            else f"{record['parent']}/{record['name']}",
            vtype=self.vtype,
            dir_type=record["dir_type"],
            owner_name=entry.owner_name,
            owner_id=entry.owner_id,
            is_latest=record["is_latest"],
            last_modified=entry.last_modified,
            version=entry.version,
            etag=entry.etag,
        )



import os

from langchain_community.embeddings import OpenAIEmbeddings
from parse import parse
from text_loaders import load_bibtex, pdf_pages

from dvcx.query import C, DatasetQuery, Object, udf
from dvcx.sql.types import Array, Float, String

# This will only run on 1/512th of the data! Change to 1 to run on all the data.
PDF_SUBSET_MOD = 512

BASE_URL = "https://proceedings.neurips.cc/paper_files/paper/"


def chomp(ss, prefix):
    assert ss.startswith(prefix)
    return ss[len(prefix) :]


@udf(
    params=(Object(load_bibtex),),
    output={
        "bib_parent": String,
        "bib_name": String,
        "bib_title": String,
        "bib_authors": Array(String),
    },
)
def get_bib_data(bib_data):
    (entry,) = bib_data.entries.keys()  # should only be one entry
    data = bib_data.entries[entry]

    url = data.fields["url"]
    url = chomp(url, BASE_URL)
    parent, name = os.path.split(url)

    title = data.fields.get("title", None)

    authors = [str(author) for author in data.persons.get("author", [])]
    return (parent, name, title, authors)


@udf(
    params=("bib_parent", "bib_name"),
    output={"hash_": String, "key": String},
)
def parse_bib_filename(parent, paper_filename):
    parse_result = parse("{}-{}.{}", paper_filename)
    hash_, _, _ = parse_result
    key_ = f"{parent}-{hash_}"
    return (hash_, key_)


@udf(
    params=("parent", "name"),
    output={
        "hash_": String,
        "object_type": String,
        "object_subtype": String,
        "ext": String,
        "key": String,
    },
)
def parse_pdf_filename(parent, filename):
    parse_result = parse("{}-{}.{}", filename)
    hash_, object_type, ext = parse_result

    # Ensure consistent camel case
    object_type = object_type.replace(" ", "").replace("_", "")

    split_list = object_type.split("-")
    assert len(split_list) <= 2
    object_type = split_list[0]
    object_subtype = split_list[1] if len(split_list) > 1 else None

    key_ = f"{parent}-{hash_}"
    return hash_, object_type, object_subtype, ext, key_


@udf(
    params=("page",),
    output={"embed": Array(Float)},
    batch=64,
)
def embed_page(pages):
    openai_api_key = os.environ["OPENAI_API_KEY"]
    assert openai_api_key.startswith("sk-")
    openai = OpenAIEmbeddings(openai_api_key=openai_api_key)
    del openai_api_key

    embed = openai.embed_documents([page[0] for page in pages])
    return [(ee,) for ee in embed]


# Remove last line to include post-2000 papers
subset_ds = (
    DatasetQuery("s3://neurips-papers/")
    .filter(~C.name.glob("*.zip"))
    .filter(~(C.parent.glob("20*/file") | C.parent.glob("20*/hash")))
)

pdf_ds = (
    subset_ds.filter(C.parent != "")
    .filter(C.name.glob("*.pdf"))
    .filter(C.random % PDF_SUBSET_MOD == 0)
    .generate(pdf_pages)
    .add_signals(parse_pdf_filename)
    .add_signals(embed_page)
)

bib_ds = (
    subset_ds.filter(C.parent != "")
    .filter((C.name == "bibtex") | C.name.glob("*.bib"))
    .add_signals(get_bib_data)
    .add_signals(parse_bib_filename)
)

pdf_ds.join(bib_ds, predicates="key")



import os

import pytest
from pydantic import Field

from dvcx.lib.feature import Feature
from dvcx.lib.feature_udf import (
    FeatureAggregator,
    FeatureBatchMapper,
    FeatureConverter,
    FeatureGenerator,
    FeatureMapper,
    OutputError,
    SchemaError,
    UserCodeError,
)
from dvcx.lib.file import File
from dvcx.query import Stream
from dvcx.sql.types import Int


class TestInput(Feature):
    name: str
    parent: str = Field(default="")
    size: int = Field(default=0)


class TestOutput(Feature):
    full_path: str
    size_squared: int = Field(default=0)


class MyAgg(FeatureAggregator):
    def __init__(self):
        super().__init__([TestInput], [TestOutput])

    def process(self, args):
        row = args[0]
        input = row[0]
        output = TestOutput(
            full_path=os.path.join(input.parent, input.name), size_squared=input.size**2
        )
        yield (output,)


def test_feature_udf_yields_values():
    name = "file.txt"
    path = "dir1/dir2"
    size = 16
    params = ((name, path, size),)

    agg = MyAgg()
    res = list(agg(params))
    assert len(res) == 1

    full_path, size_squared = res[0]
    assert full_path == os.path.join(path, name)
    assert size_squared == size**2


def test_feature_udf_returns_values():
    class MyAggReturn(FeatureAggregator):
        def __init__(self):
            super().__init__([TestInput], [TestOutput])

        def process(self, args):
            row = args[0]
            input = row[0]
            output = TestOutput(
                full_path=os.path.join(input.parent, input.name),
                size_squared=input.size**2,
            )
            return [(output,)]

    name = "f1.txt"
    path = "d1/d2/d55"
    size = 235
    params = ((name, path, size),)

    agg = MyAggReturn()
    res = list(agg(params))
    assert len(res) == 1

    full_path, size_squared = res[0]
    assert full_path == os.path.join(path, name)
    assert size_squared == size**2


class MyAggStream(FeatureAggregator):
    def __init__(self):
        super().__init__([File], [TestOutput])

    def process(self, args):
        output = TestOutput(full_path="xx", size_squared=5)
        yield (output,)


def test_feature_udf_with_stream():
    stream = File(name="tmp.jpg")
    args = ([Stream(), *list(stream._flatten())],)

    agg = MyAggStream()
    res = list(agg(args))
    assert len(res) == 1


def test_feature_udf_with_single_input_and_output():
    class MyAgg(FeatureAggregator):
        def __init__(self):
            super().__init__(TestInput, TestOutput)

        def process(self, args):
            input = args[0]
            output = TestOutput(
                full_path=os.path.join(input.parent, input.name),
                size_squared=input.size**2,
            )
            yield output

    name = "f1.txt"
    path = "dd3"
    size = 343
    params = ((name, path, size),)

    agg = MyAgg()
    res = list(agg(params))
    assert len(res) == 1

    full_path, size_squared = res[0]
    assert full_path == os.path.join(path, name)
    assert size_squared == size**2


def test_incompatible_params_types():
    with pytest.raises(SchemaError):

        class MyAggParamInstance(FeatureAggregator):
            def __init__(self):
                super().__init__(
                    (
                        ("size", Int),
                        ("test", Int),
                    ),
                    TestOutput,
                )

        MyAggParamInstance()

    with pytest.raises(SchemaError):

        class MyAggParamType(FeatureAggregator):
            def __init__(self):
                super().__init__([Int, Int], TestOutput)

        MyAggParamType()

    with pytest.raises(SchemaError):

        class MyAggOutputInstance(FeatureAggregator):
            def __init__(self):
                super().__init__([Int], TestOutput)

        MyAggOutputInstance()


def test_output_length_missmatch_with_params():
    class MyAggWrongLength(FeatureAggregator):
        def __init__(self):
            super().__init__(TestInput, [TestOutput, TestOutput])

        def process(self, args):
            input = args[0]
            output = TestOutput(
                full_path=os.path.join(input.parent, input.name),
                size_squared=input.size**2,
            )
            yield output, output, output

    with pytest.raises(OutputError):
        params = (("file", "dir", 12345),)
        agg = MyAggWrongLength()
        agg(params)


def test_output_type_missmatch_with_params_types():
    class MyAggWrongOutputType(FeatureAggregator):
        def __init__(self):
            super().__init__(TestInput, TestOutput)

        def process(self, args):
            yield 536

    with pytest.raises(OutputError):
        params = (("file", "dir", 12345),)
        agg = MyAggWrongOutputType()
        agg(params)


def test_mapper():
    class MyMap(FeatureMapper):
        def __init__(self):
            super().__init__(TestInput, TestOutput)

        def process(self, test_input):
            return TestOutput(
                full_path=os.path.join(test_input.parent, test_input.name),
                size_squared=test_input.size**2,
            )

    name = "img001.jpg"
    path = "data/dogs"
    size = 56743
    params = (name, path, size)

    agg = MyMap()
    res = agg(*params)

    full_path, size_squared = res
    assert full_path == os.path.join(path, name)
    assert size_squared == size**2


def test_mapper_stream():
    class MyMapStream(FeatureMapper):
        def __init__(self):
            super().__init__([File, TestInput], TestOutput)

        def process(self, inputs):
            stream, test_input = inputs
            return TestOutput(
                full_path=os.path.join(test_input.parent, test_input.name),
                size_squared=test_input.size**2,
            )

    stream = File(name="file.jpg")
    stream_args = (Stream(), *list(stream._flatten()))

    name = "img001.jpg"
    path = "data/dogs"
    size = 56743
    args = (*stream_args, name, path, size)

    agg = MyMapStream()
    res = agg(*args)

    full_path, size_squared = res
    assert full_path == os.path.join(path, name)
    assert size_squared == size**2


def test_batch_mapper():
    class MyBatchMap(FeatureBatchMapper):
        def __init__(self):
            super().__init__(TestInput, TestOutput, 2)

        def process(self, test_inputs):
            for test_input in test_inputs:
                output = TestOutput(
                    full_path=os.path.join(test_input.parent, test_input.name),
                    size_squared=test_input.size**2,
                )
                yield output

    name = "img001.jpg"
    path = "data/dogs"
    size = 56743
    params = ((name, path, size), ("1_" + name, "1_" + path, size * 2))

    agg = MyBatchMap()
    res = agg(params)

    assert len(res) == 2

    full_path, size_squared = res[0]
    assert full_path == os.path.join(path, name)
    assert size_squared == size**2

    full_path, size_squared = res[1]
    assert full_path == os.path.join("1_" + path, "1_" + name)
    assert size_squared == (size * 2) ** 2


def test_generator():
    class MyGen(FeatureGenerator):
        def __init__(self):
            super().__init__(TestInput, TestOutput)

        def process(self, test_input):
            yield TestOutput(
                full_path=os.path.join(test_input.parent, test_input.name) + "_1",
                size_squared=test_input.size**2,
            )

            yield TestOutput(
                full_path=os.path.join(test_input.parent, test_input.name) + "_2",
                size_squared=test_input.size**3,
            )

    name = "img742.jpg"
    path = "d1/cats"
    size = 384
    params = (name, path, size)

    agg = MyGen()
    res = agg(*params)

    assert len(res) == 2

    full_path, size_squared = res[0]
    assert full_path == os.path.join(path, name) + "_1"
    assert size_squared == size**2

    full_path, size_squared = res[1]
    assert full_path == os.path.join(path, name) + "_2"
    assert size_squared == size**3


def test_feature_udf_output_value_instead_of_list():
    class MyAggOutputValueInsteadOfList(FeatureMapper):
        def __init__(self):
            super().__init__([TestInput], [TestOutput])

        def process(self, args):
            input = args[0]
            return TestOutput(
                full_path=os.path.join(input.parent, input.name),
                size_squared=input.size**2,
            )

    name = "file.txt"
    path = "dir1/dir2"
    size = 16
    params = (name, path, size)

    agg = MyAggOutputValueInsteadOfList()
    with pytest.raises(OutputError):
        list(agg(*params))


def test_feature_udf_output_wrong_type():
    class MyAggWrongOutputType(FeatureMapper):
        def __init__(self):
            super().__init__([TestInput], TestOutput)

        def process(self, args):
            return 34

    name = "file.txt"
    path = "dir1/dir2"
    size = 16
    params = (name, path, size)

    agg = MyAggWrongOutputType()
    with pytest.raises(OutputError):
        list(agg(*params))


def test_error_in_user_code():
    class MyBuggyClass(FeatureMapper):
        def __init__(self):
            super().__init__(TestInput, TestOutput)

        def process(self, args):
            return 3.14 / 0

    name = "file.txt"
    path = "dir1/dir2"
    size = 16
    params = (name, path, size)

    agg = MyBuggyClass()
    with pytest.raises(UserCodeError):
        list(agg(*params))


def test_split():
    features = [File]
    f1 = File(name="abc")._flatten()
    f2 = File(name="zxc")._flatten()

    rows = [f1, f2]
    clean_rows, streams = FeatureConverter._separate_streams_from_rows(rows, features)
    assert clean_rows == rows
    assert streams == [None, None]

    rows_with_streams = [(Stream(), *f1), (Stream(), *f2)]
    clean_rows, streams = FeatureConverter._separate_streams_from_rows(
        rows_with_streams, features
    )
    assert clean_rows == rows
    assert len(streams) == 2
    assert isinstance(streams[0], Stream)
    assert isinstance(streams[1], Stream)



import base64
import pickle
from abc import abstractmethod
from collections.abc import Callable
from typing import Any


class Serializable:
    @abstractmethod
    def clone_params(self) -> tuple[Callable[..., Any], list[Any], dict[str, Any]]:
        """
        Returns the class, args, and kwargs needed to instantiate a cloned copy
        of this instance for use in separate processes or machines.
        """

    def serialize(self) -> str:
        """
        Returns a string representation of clone params.
        This is useful for storing the state of an object in environment variable.
        """
        return base64.b64encode(pickle.dumps(self.clone_params())).decode()


def deserialize(s: str) -> Serializable:
    """
    Returns a new instance of the class represented by the string.
    """
    (f, args, kwargs) = pickle.loads(base64.b64decode(s.encode()))  # noqa: S301
    return f(*args, **kwargs)



from datetime import datetime, timezone

from dvcx.node import Entry

utc = timezone.utc
TIME_ZERO = datetime.fromtimestamp(0, tz=utc)

ENTRIES = [
    Entry.from_dir(
        parent="",
        name="cats",
        last_modified=TIME_ZERO,
        etag="",
        version="",
        is_latest=True,
        size=0,
        owner_name="",
        owner_id="",
    ),
    Entry.from_dir(
        parent="",
        name="dogs",
        last_modified=TIME_ZERO,
        etag="",
        version="",
        is_latest=True,
        size=0,
        owner_name="",
        owner_id="",
    ),
    Entry.from_file(
        parent="",
        name="description",
        etag="60a7605e934638ab9113e0f9cf852239",
        version="7e589b7d-382c-49a5-931f-2b999c930c5e",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=13,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
    Entry.from_file(
        parent="cats",
        name="cat1",
        etag="4a4be40c96ac6314e91d93f38043a634",
        version="309eb4a4-bba9-47c1-afcd-d7c51110af6f",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=4,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
    Entry.from_file(
        parent="cats",
        name="cat2",
        etag="0268c692ff940a830e1e7296aa48c176",
        version="f9d168d3-6d1b-47ef-8f6a-81fce48de141",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=4,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
    Entry.from_dir(
        parent="dogs",
        name="others",
        last_modified=TIME_ZERO,
        etag="",
        version="",
        is_latest=True,
        size=0,
        owner_name="",
        owner_id="",
    ),
    Entry.from_file(
        parent="dogs",
        name="dog1",
        etag="8fdb60801e9d39a5286aa01dd1f4f4f3",
        version="b9c31cf7-d011-466a-bf16-cf9da0cb422a",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=4,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
    Entry.from_file(
        parent="dogs",
        name="dog2",
        etag="2d50c921b22aa164a56c68d71eeb4100",
        version="3a8bb6d9-38db-47a8-8bcb-8972ea95aa20",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=3,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
    Entry.from_file(
        parent="dogs",
        name="dog3",
        etag="33c6c2397a1b079e903c474df792d0e2",
        version="ee49e963-36a8-492a-b03a-e801b93afb40",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=4,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
    Entry.from_file(
        parent="dogs/others",
        name="dog4",
        etag="a5e1a5d93ff242b745f5cf87aeb726d5",
        version="c5969421-6900-4060-bc39-d54f4a49b9fc",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=4,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
]

# files with directory name collisions:
#   dogs/others/
#   dogs/others
#   dogs/
INVALID_ENTRIES = [
    Entry.from_file(
        parent="dogs/others",
        name="",
        etag="68b329da9893e34099c7d8ad5cb9c940",
        version="85969421-6900-4060-bc39-d54f4a49b9ab",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=4,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
    Entry.from_file(
        parent="dogs",
        name="others",
        etag="68b329da9893e34099c7d8ad5cb9c940",
        version="85969421-6900-4060-bc39-d54f4a49b9ab",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=4,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
    Entry.from_file(
        parent="dogs",
        name="",
        etag="68b329da9893e34099c7d8ad5cb9c940",
        version="85969421-6900-4060-bc39-d54f4a49b9ab",
        is_latest=True,
        last_modified=datetime(2023, 2, 27, 18, 28, 54, tzinfo=utc),
        size=4,
        owner_name="webfile",
        owner_id="75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a",
    ),
]



import pytest

from dvcx.sql import literal, select
from dvcx.sql.functions import string


def test_length(warehouse):
    query = select(string.length(literal("abcdefg")))
    result = tuple(warehouse.db.execute(query))
    assert result == ((7,),)


@pytest.mark.parametrize(
    "args,expected",
    [
        ([literal("abc//def/g/hi"), literal("/")], ["abc", "", "def", "g", "hi"]),
        ([literal("abc//def/g/hi"), literal("/"), 2], ["abc", "", "def/g/hi"]),
    ],
)
def test_split(warehouse, args, expected):
    query = select(string.split(*args))
    result = tuple(warehouse.dataset_rows_select(query))
    assert result == ((expected,),)



import inspect
import io
import sys
import traceback
from collections.abc import Sequence
from typing import Optional, Union

from dvcx.catalog import Catalog
from dvcx.lib.feature import Feature
from dvcx.lib.feature_types import FeatureLike, FeatureTypes
from dvcx.lib.file import File
from dvcx.lib.udf import Aggregator, BatchMapper, Generator, Mapper, UDFBase
from dvcx.lib.utils import DvcxError
from dvcx.query import Stream


class ValidationError(DvcxError):
    pass


class SchemaError(ValidationError):
    def __init__(self, udf_name: str, context: str, message: str):
        super().__init__(f"'{udf_name}' {context} schema validation error: {message}")


class OutputError(ValidationError):
    def __init__(self, udf_name: str, message: str, num: Optional[int] = None):
        num_str = "" if num is None else f"#{num + 1} "
        super().__init__(f"Output {num_str}of '{udf_name}' error: {message}")


class UserCodeError(DvcxError):
    def __init__(self, class_name: str, message: str):
        super().__init__(f"Error in user code in class '{class_name}': {message}")


class FeatureConverter:
    @property
    def udf_params_list(self):
        return self._udf_params_list

    @property
    def udf_output_spec(self):
        return self._udf_output_spec

    @property
    def cache(self):
        return self._udf.catalog.cache

    @staticmethod
    def has_feature_stream(fr_classes: Sequence[type[Feature]]):
        return any(
            f._is_file  # type: ignore[attr-defined]
            for f in fr_classes
        )

    @staticmethod
    def has_row_stream(row):
        if len(row) == 0:
            return False
        return isinstance(row[0], (Stream, io.IOBase))

    def __init__(
        self,
        udf: UDFBase,
        inputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        outputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
    ):
        self._udf = udf

        self._inputs, self._is_single_input = self._convert_to_sequence(inputs)
        self._outputs, self._is_single_output = self._convert_to_sequence(outputs)

        self._validate_schema("params", self._inputs)
        self._validate_schema("output", self._outputs)

        self._udf_params_list = self.get_flattened_params(self._inputs)
        self._udf_output_spec = Feature._features_to_udf_spec(self._outputs)  # type: ignore[attr-defined]

    @staticmethod
    def get_flattened_params(fr_classes: Sequence[type[Feature]]):
        udf_params_spec = Feature._features_to_udf_spec(fr_classes)
        stream_prm = (
            [Stream()] if FeatureConverter.has_feature_stream(fr_classes) else []
        )
        return stream_prm + list(udf_params_spec.keys())

    @staticmethod
    def _convert_to_sequence(
        arg: Union[FeatureLike, Sequence[FeatureLike]],
    ) -> tuple[Sequence[type[Feature]], bool]:
        if not isinstance(arg, Sequence):
            return FeatureTypes._to_features(*[arg]), True
        return FeatureTypes._to_features(*arg), False

    @staticmethod
    def deserialize(
        rows: Sequence[Sequence],
        params: Sequence[str],
        fr_classes: Sequence[type[Feature]],
        catalog: Catalog,
        caching_enabled: bool,
    ) -> Sequence[Sequence[Feature]]:
        clean_rows, streams = FeatureConverter._separate_streams_from_rows(
            rows, fr_classes
        )

        feature_rows = [
            FeatureConverter._row_with_params_to_features(row, fr_classes, params)
            for row in clean_rows
        ]

        for features, stream in zip(feature_rows, streams):
            for feature in features:
                if isinstance(feature, File):
                    feature.set_catalog(catalog)  # type: ignore [attr-defined]
                    feature.set_file(stream, caching_enabled)  # type: ignore [attr-defined]

        return feature_rows

    @staticmethod
    def _separate_streams_from_rows(
        rows, fr_classes: Sequence[type[Feature]]
    ) -> tuple[Sequence, Sequence]:
        streams = []
        res_rows = []
        if FeatureConverter.has_feature_stream(fr_classes):
            for row in rows:
                if FeatureConverter.has_row_stream(row):
                    streams.append(row[0])
                    res_rows.append(row[1:])
                else:
                    streams.append(None)  # type: ignore [arg-type]
                    res_rows.append(row)
        else:
            res_rows = rows
        return res_rows, streams

    @staticmethod
    def _row_with_params_to_features(
        row: Sequence, fr_classes: Sequence[type[Feature]], params: Sequence[str]
    ) -> Sequence[Feature]:
        new_params = (
            params
            if not FeatureConverter.has_feature_stream(fr_classes)
            else params[1:]
        )
        return [cls._unflatten(dict(zip(new_params, row))) for cls in fr_classes]

    def _validate_schema(self, context: str, features: Sequence[type[Feature]]):
        for type_ in features:
            if not isinstance(type_, type):
                raise SchemaError(
                    self._udf.name,
                    context,
                    "cannot accept objects, a 'Feature' class must be provided",
                )
            if not issubclass(type_, Feature):
                raise SchemaError(
                    self._udf.name,
                    context,
                    f"cannot accept type '{type_.__name__}', "
                    f"the type must be a subclass of 'Feature'",
                )

    def validate_output_obj(self, result_objs, *args, **kwargs):
        for row in result_objs:
            if not isinstance(row, (list, tuple)) and isinstance(
                self._outputs, (list, tuple)
            ):
                raise OutputError(
                    self._udf.name,
                    f"expected list of objects, "
                    f"but found a single value of type '{type(row).__name__}'",
                )

            if len(row) != len(self._outputs):
                raise OutputError(
                    self._udf.name,
                    f"length mismatch - expected {len(self._outputs)} "
                    f"objects, but {len(row)} were provided",
                )

            for num, (o, type_) in enumerate(zip(row, self._outputs)):
                if not isinstance(o, type_):
                    raise OutputError(
                        self._udf.name,
                        f"expected type '{type_.__name__}',"
                        f" but found type '{type(o).__name__}'",
                        num=num,
                    )

    def process_rows(self, rows, is_input_batched=True, is_output_batched=True):
        obj_rows = FeatureConverter.deserialize(
            rows,
            self._udf.params,
            self._inputs,
            self._udf.catalog,
            self._udf.caching_enabled,
        )

        if self._is_single_input:
            obj_rows = [objs[0] for objs in obj_rows]

        if not is_input_batched:
            assert (
                len(obj_rows) == 1
            ), f"{self._udf.name} takes {len(obj_rows)} rows while it's not batched"
            obj_rows = obj_rows[0]

        result_objs = self.run_user_code(obj_rows)

        if not is_output_batched:
            result_objs = [result_objs]

        if self._is_single_output:
            result_objs = [[x] for x in result_objs]

        self.validate_output_obj(result_objs)

        res = [Feature._flatten_list(objs) for objs in result_objs]

        if not is_output_batched:
            assert len(res) == 1, (
                f"{self._udf.name} returns {len(obj_rows)} "
                f"rows while it's not batched"
            )
            res = res[0]
        return res

    def run_user_code(self, obj_rows):
        try:
            result_objs = self._udf.process(obj_rows)
            if inspect.isgeneratorfunction(self._udf.process):
                result_objs = list(result_objs)
        except Exception as e:  # noqa: BLE001
            msg = (
                f"============== Error in user code: '{self._udf.name}' =============="
            )
            print(msg)
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback.tb_next)
            print("=" * len(msg))
            raise UserCodeError(self._udf.name, str(e)) from None
        return result_objs


class FeatureAggregator(Aggregator):
    def __init__(
        self,
        inputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        outputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        batch=1,
    ):
        self._fc = FeatureConverter(self, inputs, outputs)
        super().__init__(self._fc.udf_params_list, self._fc.udf_output_spec, batch)

    def __call__(self, rows):
        return self._fc.process_rows(rows)


class FeatureMapper(Mapper):
    def __init__(
        self,
        inputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        outputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        batch=1,
    ):
        self._fc = FeatureConverter(self, inputs, outputs)
        super().__init__(self._fc.udf_params_list, self._fc.udf_output_spec, batch)

    def __call__(self, *row):
        return self._fc.process_rows([row], False, False)


class FeatureBatchMapper(BatchMapper):
    def __init__(
        self,
        inputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        outputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        batch=1,
    ):
        self._fc = FeatureConverter(self, inputs, outputs)
        super().__init__(self._fc.udf_params_list, self._fc.udf_output_spec, batch)

    def __call__(self, rows):
        return self._fc.process_rows(rows)


class FeatureGenerator(Generator):
    def __init__(
        self,
        inputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        outputs: Union[FeatureLike, Sequence[FeatureLike]] = (),
        batch=1,
    ):
        self._fc = FeatureConverter(self, inputs, outputs)
        super().__init__(self._fc.udf_params_list, self._fc.udf_output_spec, batch)

    def __call__(self, *row):
        return self._fc.process_rows([row], False)



"""
fashion-clip UDF, using the
[fashion-clip](https://pypi.org/project/fashion-clip/#description) package.

Generated embeddings are stored json-encoded.

To install script dependencies: pip install tabulate fashion-clip
"""

import json

from fashion_clip.fashion_clip import FashionCLIP
from sqlalchemy import JSON
from tabulate import tabulate

from dvcx.lib.param import Image
from dvcx.query import C, DatasetQuery, udf


@udf(
    params=(Image(),),
    output={"fclip": JSON},
    method="fashion_clip",
    batch=10,
)
class MyFashionClip:
    def __init__(self):
        self.fclip = FashionCLIP("fashion-clip")

    def fashion_clip(self, inputs):
        embeddings = self.fclip.encode_images(
            [input[0] for input in inputs], batch_size=1
        )
        return [(json.dumps(emb),) for emb in embeddings.tolist()]


if __name__ == "__main__":
    # This example processes 5 objects in the new dataset and generates the
    # embeddings for them.
    DatasetQuery(path="gcs://dvcx-zalando-hd-resized/zalando-hd-resized/").filter(
        C.name.glob("*.jpg")
    ).limit(5).add_signals(MyFashionClip).save("zalando_hd_emb")

    print(tabulate(DatasetQuery(name="zalando_hd_emb").results()[:5]))



import logging
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, ClassVar, Optional, Union

import sqlalchemy as sa
from attrs import frozen
from sqlalchemy.sql.roles import DDLRole

from dvcx.data_storage.serializer import Serializable

if TYPE_CHECKING:
    from sqlalchemy import MetaData, Table
    from sqlalchemy.engine.base import Engine
    from sqlalchemy.engine.interfaces import Dialect
    from sqlalchemy.sql.compiler import Compiled
    from sqlalchemy.sql.elements import ClauseElement


logger = logging.getLogger("dvcx")

RANDOM_BITS = 63  # size of the random integer field

SELECT_BATCH_SIZE = 100_000  # number of rows to fetch at a time


@frozen
class DatabaseEngine(ABC, Serializable):
    dialect: ClassVar["Dialect"]

    engine: "Engine"
    metadata: "MetaData"

    @abstractmethod
    def clone(self) -> "DatabaseEngine":
        """Clones DatabaseEngine implementation."""

    @classmethod
    def compile(cls, statement: "ClauseElement", **kwargs) -> "Compiled":
        """
        Compile a sqlalchemy query or ddl object to a Compiled object.

        Use the `string` and `params` properties of this object to get
        the resulting sql string and parameters.
        """
        if not isinstance(statement, DDLRole):
            # render_postcompile is needed for in_ queries to work
            kwargs["compile_kwargs"] = {
                **kwargs.pop("compile_kwargs", {}),
                "render_postcompile": True,
            }
        return statement.compile(dialect=cls.dialect, **kwargs)

    @classmethod
    def compile_to_args(
        cls, statement: "ClauseElement", **kwargs
    ) -> Union[tuple[str], tuple[str, dict[str, Any]]]:
        """
        Compile a sqlalchemy query or ddl object to an args tuple.

        This tuple is formatted specifically for calling
        `cursor.execute(*args)` according to the python DB-API.
        """
        result = cls.compile(statement, **kwargs)
        params = result.params
        if params is None:
            return (result.string,)
        return result.string, params

    @abstractmethod
    def execute(
        self,
        query,
        cursor: Optional[Any] = None,
        conn: Optional[Any] = None,
    ) -> Iterator[tuple[Any, ...]]: ...

    @abstractmethod
    def executemany(
        self, query, params, cursor: Optional[Any] = None
    ) -> Iterator[tuple[Any, ...]]: ...

    @abstractmethod
    def execute_str(self, sql: str, parameters=None) -> Iterator[tuple[Any, ...]]: ...

    @abstractmethod
    def insert_dataframe(self, table_name: str, df) -> int: ...

    @abstractmethod
    def close(self) -> None: ...

    @abstractmethod
    def transaction(self): ...

    def has_table(self, name: str) -> bool:
        """
        Return True if a table exists with the given name
        """
        return sa.inspect(self.engine).has_table(name)

    @abstractmethod
    def create_table(self, table: "Table", if_not_exists: bool = True) -> None: ...

    @abstractmethod
    def drop_table(self, table: "Table", if_exists: bool = False) -> None: ...

    @abstractmethod
    def rename_table(self, old_name: str, new_name: str): ...



from dvcx.query import C, DatasetQuery, udf
from dvcx.sql import literal
from dvcx.sql.functions import array, greatest, least, path, string
from dvcx.sql.types import Array, String


@udf(
    params=(C.name,),
    output={"num_chars": Array(String)},
)
def num_chars_udf(name):
    parts = name.split(".")
    if len(parts) > 1:
        return (list(parts[1]),)
    return ([],)


def show(dataset):
    print(*dataset.results(), sep="\n", end="\n\n")


ds = DatasetQuery("gcs://dvcx-datalakes/dogs-and-cats/", anon=True)
show(ds.limit(5).add_signals(num_chars_udf).select(C.name, C.num_chars))
show(
    ds.limit(5).select(
        C.name,
        string.length(C.name),
        string.split(C.name, literal(".")),
    )
)
show(
    ds.limit(5).select(
        C.name,
        path.file_stem(C.name),
        path.file_ext(C.name),
    )
)
show(
    ds.limit(10)
    .mutate(
        a=array.length(string.split(C.parent, literal("/"))),
        b=array.length(string.split(C.name, literal("0"))),
    )
    .select(C.a, C.b, greatest(C.a, C.b), least(C.a, C.b))
)


"""
Expected output:
('', [])
('cat.1.jpg', ['1'])
('cat.1.json', ['1'])
('cat.10.jpg', ['1', '0'])
('cat.10.json', ['1', '0'])

('', 0, [''])
('cat.1.jpg', 9, ['cat', '1', 'jpg'])
('cat.1.json', 10, ['cat', '1', 'json'])
('cat.10.jpg', 10, ['cat', '10', 'jpg'])
('cat.10.json', 11, ['cat', '10', 'json'])

('', '', '')
('cat.1.jpg', 'cat.1', 'jpg')
('cat.1.json', 'cat.1', 'json')
('cat.10.jpg', 'cat.10', 'jpg')
('cat.10.json', 'cat.10', 'json')

(3, 1, 3, 1)
(3, 1, 3, 1)
(3, 1, 3, 1)
(3, 2, 3, 2)
(3, 2, 3, 2)
(3, 3, 3, 3)
(3, 3, 3, 3)
(3, 4, 4, 3)
(3, 4, 4, 3)
(3, 3, 3, 3)
"""



from collections.abc import Iterable

from dvcx.node import DirType, NodeWithPath


class DataSource:
    def __init__(self, listing, node, as_container=False):
        self.listing = listing
        self.node = node
        self.as_container = (
            as_container  # Indicates whether a .tar file is handled as a container
        )

    def get_full_path(self):
        return self.get_node_full_path(self.node)

    def get_node_full_path(self, node):
        return self.listing.client.get_full_path(node.full_path)

    def get_node_full_path_from_path(self, full_path):
        return self.listing.client.get_full_path(full_path)

    def is_single_object(self):
        return self.node.dir_type == DirType.FILE or (
            not self.as_container and self.node.dir_type == DirType.TAR_ARCHIVE
        )

    def is_container(self):
        return not self.is_single_object()

    def ls(self, fields) -> Iterable[tuple]:
        if self.is_single_object():
            return [tuple(getattr(self.node, f) for f in fields)]
        return self.listing.ls_path(self.node, fields)

    def dirname(self):
        if self.is_single_object():
            return self.node.parent
        return self.node.path

    def find(self, *, sort=None):
        if self.is_single_object():
            return [NodeWithPath(self.node, [])]

        return self.listing.subtree_files(self.node, sort=sort)



import sys
from pathlib import Path

import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st

from dvcx.client.local import FileClient
from tests.utils import uppercase_scheme

non_null_text = st.text(
    alphabet=st.characters(blacklist_categories=["Cc", "Cs"]), min_size=1
)


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture], deadline=None)
@given(rel_path=non_null_text)
def test_parse_url(cloud_test_catalog, rel_path, cloud_type):
    if cloud_type == "file":
        assume(not rel_path.startswith("/"))
    bucket_uri = cloud_test_catalog.src_uri
    url = f"{bucket_uri}/{rel_path}"
    catalog = cloud_test_catalog.catalog
    client, rel_part = catalog.parse_url(url)
    if cloud_type == "file":
        root_uri = FileClient.root_path().as_uri()
        assert client.uri == root_uri
        assert rel_part == url[len(root_uri) :]
    else:
        assert client.uri == bucket_uri
        assert rel_part == rel_path


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture], deadline=None)
@given(rel_path=non_null_text)
def test_parse_url_uppercase_scheme(cloud_test_catalog, rel_path, cloud_type):
    if cloud_type == "file":
        assume(not rel_path.startswith("/"))
    bucket_uri = cloud_test_catalog.src_uri
    bucket_uri_upper = uppercase_scheme(bucket_uri)
    url = f"{bucket_uri_upper}/{rel_path}"
    catalog = cloud_test_catalog.catalog
    client, rel_part = catalog.parse_url(url)
    if cloud_type == "file":
        root_uri = FileClient.root_path().as_uri()
        assert client.uri == root_uri
        assert rel_part == url[len(root_uri) :]
    else:
        assert client.uri == bucket_uri
        assert rel_part == rel_path


@pytest.mark.parametrize("cloud_type", ["file"], indirect=True)
def test_parse_file_absolute_path_without_protocol(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog
    working_dir = Path().absolute()
    root_uri = FileClient.root_path().as_uri()
    client, rel_part = catalog.parse_url(str(working_dir / Path("animals")))
    working_dir = Path().absolute()
    root_uri = FileClient.root_path().as_uri()
    assert client.uri == root_uri
    assert rel_part == (working_dir / Path("animals")).as_uri()[len(root_uri) :]


@pytest.mark.parametrize("cloud_type", ["file"], indirect=True)
def test_parse_file_relative_path_multiple_dirs_back(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog
    client, rel_part = catalog.parse_url("../../animals")
    root_uri = FileClient.root_path().as_uri()
    assert client.uri == root_uri
    assert (
        rel_part
        == (Path().absolute().parents[1] / Path("animals")).as_uri()[len(root_uri) :]
    )


@pytest.mark.parametrize("cloud_type", ["file"], indirect=True)
@pytest.mark.parametrize("url", ["./animals", "animals"])
def test_parse_file_relative_path_working_dir(cloud_test_catalog, url):
    catalog = cloud_test_catalog.catalog
    client, rel_part = catalog.parse_url(url)
    root_uri = FileClient.root_path().as_uri()
    assert client.uri == root_uri
    assert rel_part == (Path().absolute() / Path("animals")).as_uri()[len(root_uri) :]


@pytest.mark.parametrize("cloud_type", ["file"], indirect=True)
def test_parse_file_relative_path_home_dir(cloud_test_catalog):
    if sys.platform == "win32":
        # home dir shortcut is not available on windows
        pytest.skip()
    catalog = cloud_test_catalog.catalog
    client, rel_part = catalog.parse_url("~/animals")
    root_uri = FileClient.root_path().as_uri()
    assert client.uri == root_uri
    assert rel_part == (Path().home() / Path("animals")).as_uri()[len(root_uri) :]



from PIL import Image
from torch import Tensor
from torchvision.transforms import ToTensor

from dvcx.lib.file import File
from dvcx.lib.image import ImageReader, convert_image


def test_convert_image(tmp_path):
    file_name = "img.jpg"
    file_path = tmp_path / file_name

    img = Image.new(mode="RGB", size=(64, 64))
    img.save(file_path)

    converted_img = convert_image(
        img,
        mode="RGBA",
        size=(32, 32),
        transform=ToTensor(),
    )
    assert isinstance(converted_img, Tensor)
    assert converted_img.size() == (4, 32, 32)


def test_image_file_reader(tmp_path, catalog, mocker):
    file_name = "img.jpg"
    file_path = tmp_path / file_name

    img = Image.new(mode="RGB", size=(64, 64))
    img.save(file_path)

    kwargs = {
        "mode": "RGBA",
        "size": (32, 32),
        "transform": ToTensor(),
        "open_clip_model": None,
    }
    reader = ImageReader(**kwargs)
    file = File(name=file_name, source=f"file://{tmp_path}")
    file.set_catalog(catalog)

    convert_image = mocker.patch("dvcx.lib.image.convert_image")

    with open(file_path, "rb") as fd:
        file.set_file(fd, caching_enabled=False)
        reader(file.get_value())
        assert len(convert_image.call_args.args) == 1
        assert isinstance(convert_image.call_args.args[0], Image.Image)
        assert convert_image.call_args.kwargs == kwargs



import os
import shutil
import tempfile
from abc import ABC
from contextlib import AbstractContextManager

from dvcx.cache import UniqueId


class AbstractCachedStream(AbstractContextManager, ABC):
    def __init__(self, stream, size, catalog, uid: UniqueId):
        self.stream = stream
        self.size = size
        self.catalog = catalog
        self.uid = uid
        self.mode = "rb"

    def set_mode(self, mode):
        self.mode = mode


class ProgressiveCacheStream(AbstractCachedStream):
    BUF_SIZE = 4096

    def __init__(self, stream, size, catalog, uid: UniqueId):
        super().__init__(stream, size, catalog, uid)

        self.target_path = self.catalog.cache.path_from_checksum(self.uid.get_hash())
        self.cached_file = None

        self.temp_file = None
        self.temp_file_pos = 0

    def __enter__(self):
        if os.path.exists(self.target_path):
            self.cached_file = open(self.target_path, mode=self.mode)  # noqa: SIM115
            return self.cached_file

        tmp_dir = self.catalog.cache.tmp_dir
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
        self.temp_file = tempfile.NamedTemporaryFile(
            prefix=str(self.uid.get_hash()), dir=tmp_dir, delete=False
        )
        return self

    def __exit__(self, *args):
        self.close()

    def read(self, size=-1):
        buf = self.stream.read(size)
        pos = self.stream.tell()

        if pos >= self.temp_file_pos:
            self._cache_catch_up(pos, buf)

        return buf

    def close(self):
        if self.cached_file:
            self.cached_file.close()

        if self.temp_file:
            if self.temp_file_pos < self.size:
                self._cache_catch_up(self.size)

            self.temp_file.close()
            if not os.path.exists(self.target_path):
                os.makedirs(os.path.dirname(self.target_path), exist_ok=True)
                shutil.move(self.temp_file.name, self.target_path)

            self.stream.close()

    def _cache_catch_up(self, pos_target, latest_buf=None):
        pos_to_restore = self.stream.tell()
        try:
            remainder = pos_target - self.temp_file_pos
            self.stream.seek(self.temp_file_pos)
            while remainder > 0:
                chunk_size = min(self.BUF_SIZE, remainder)
                buf = self.stream.read(chunk_size)
                self._cache_update(buf)
                remainder -= len(buf)
        finally:
            self.stream.seek(pos_to_restore)

    def _cache_update(self, buf):
        length = len(buf)
        self.temp_file.write(buf)
        self.temp_file_pos += length

    def seek(self, offset, whence=0):
        return self.stream.seek(offset, whence)

    def tell(self):
        return self.stream.tell()


class PreCachedStream(AbstractCachedStream):
    def __init__(self, stream, size, catalog, uid: UniqueId):
        super().__init__(stream, size, catalog, uid)
        self.client = self.catalog.get_client(self.uid.storage)
        self.cached_file = None

    def get_path_in_cache(self):
        return self.catalog.cache.path_from_checksum(self.uid.get_hash())

    def __enter__(self):
        self.client.download(self.uid)
        self.cached_file = open(self.get_path_in_cache(), self.mode)  # noqa: SIM115
        return self.cached_file

    def __exit__(self, *args):
        self.cached_file.close()


class PreDownloadStream(PreCachedStream):
    def __exit__(self, *args):
        super().__exit__(*args)
        self.catalog.cache.remove(self.uid)



import inspect
import types
from typing import TYPE_CHECKING, Callable

from dvcx.catalog import get_catalog
from dvcx.query import udf
from dvcx.query.schema import DatasetRow
from dvcx.sql.types import Int, String

if TYPE_CHECKING:
    from dvxc.query.udf import UDFWrapper
    from typing_extensions import Self


class UDFBase:
    DEF_OUTPUT_NAME = "result"
    DEF_OUTPUT_TYPE = String

    def __init__(self, params=None, output=None, batch=1):
        sign_params, sign_output = self._get_signature(self.process)

        self._params = params or sign_params
        self._output = output or sign_output
        self._batch = batch
        self._caching_enabled = False
        self._catalog = get_catalog().copy(db=False)

    @property
    def params(self):
        return self._params

    @property
    def output(self):
        return self._output

    @property
    def batch(self):
        return self._batch

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def caching_enabled(self):
        return self._caching_enabled

    def enable_caching(self):
        self._caching_enabled = True

    def set_catalog(self, catalog):
        self._catalog = catalog.copy(db=False)

    @property
    def catalog(self):
        return self._catalog

    def to_udf_wrapper(self) -> "UDFWrapper":
        udf_wrapper = udf(params=self.params, output=self.output, batch=self.batch)
        return udf_wrapper(self)

    def bootstrap(self):
        """Initialization process executed on each worker before processing begins.
        This is needed for tasks like pre-loading ML models prior to scoring.
        """

    def teardown(self):
        """Teardown process executed on each process/worker after processing ends.
        This is needed for tasks like closing connections to end-points.
        """

    def process(self, *args, **kwargs):
        """Abstract processing method that needs to be re-defined in child classes."""
        NotImplementedError(f"UDF processing is not implemented in class {self.name}")

    def validate_results(self, results, *args, **kwargs):
        return results

    def __call__(self, *args, **kwargs):
        results = self.process(*args, **kwargs)
        return self.validate_results(results, *args, **kwargs)

    @staticmethod
    def _get_signature(func):
        sign = inspect.signature(func)
        params = tuple(sign.parameters.keys())
        name = func.__name__ if inspect.isfunction(func) else func.__class__.__name__
        output = {
            UDFBase.DEF_OUTPUT_NAME: UDFBase._get_result_type(
                name, sign.return_annotation
            )
        }
        return params, output

    @staticmethod
    def _get_result_type(func_name, return_annotation):
        if return_annotation is inspect.Signature.empty:
            return UDFBase.DEF_OUTPUT_TYPE
        if return_annotation is str:
            return String
        if return_annotation is int:
            return Int
        raise ValueError(
            f"UDF '{func_name}' output type must be str or int"
            f" while {return_annotation} was provided"
        )

    def _validate_result_type(self, results, *args):
        if not isinstance(results, tuple):
            raise TypeError(
                f"{self.name} returned {type(results)} values"
                f" while tuple was expected. Inputs: {args} Result: {results}"
            )

    def _validate_result_length(self, results, *args):
        out_len = len(self.output)
        res_len = len(results)

        if out_len != res_len:
            raise ValueError(
                f"{self.name} returned {res_len} values"
                f" while {out_len} expected. Inputs: {args} Result: {results}"
            )

    def _validate_result_batch_length(self, results, *args):
        out_len = len(self.output)

        for res in results:
            if not isinstance(res, tuple):
                raise TypeError(
                    f"{self.name} must return a sequence of tuples."
                    f" This may occur if using 'return' instead of 'yield'."
                    f" Inputs: {args} Result: {res}"
                )

            res_len = len(res)
            if out_len != res_len:
                raise ValueError(
                    f"{self.name} returned {res_len} values"
                    f" while {out_len} expected. Inputs: {args} Result: {res}"
                )

    @classmethod
    def _from_func(cls, func: Callable, params=None, output=None, batch=1) -> "Self":
        sign_params, sign_output = UDFBase._get_signature(func)
        udf = cls(params or sign_params, output or sign_output, batch=batch)
        udf.process = func  # type: ignore[method-assign]
        return udf

    @classmethod
    def from_func(
        cls, func: Callable, params=None, output=None, batch: int = 1
    ) -> "UDFBase":
        return cls._from_func(func, params, output, batch)


class Mapper(UDFBase):
    def __init__(self, params=None, output=None, batch=1):
        if batch > 1:
            raise ValueError("Mapper does not support batch")
        super().__init__(params, output)

    def validate_results(self, results, *args, **kwargs):
        self._validate_result_type(results, *args)
        self._validate_result_length(results, *args)
        return results

    @classmethod
    def from_func(cls, func, params=None, output=None, batch=1) -> "Mapper":
        if batch == 1:
            return super()._from_func(func, params, output)
        return super(Mapper, BatchMapper)._from_func(func, params, output, batch)


class BatchMapper(Mapper):
    def __init__(self, params=None, output=None, batch=1000):
        if batch == 1:
            raise ValueError(f"{self.name} must be batch UDF")
        UDFBase.__init__(self, params, output, batch)

    def validate_results(self, results, *args, **kwargs):
        # ToDo: this should be done in the level above - together with serialization
        # to not to create the whole copy
        if isinstance(results, types.GeneratorType):
            results = tuple(results)

        self._validate_result_type(results, *args)
        self._validate_result_batch_length(results, *args)
        return results


class Generator(UDFBase):
    def __init__(self, params=None, output=None, batch=1):
        if batch > 1:
            raise ValueError("Generator does not support batch")

        output = output or {}
        super().__init__(params, output=DatasetRow.schema | output)

    def validate_results(self, results, *args, **kwargs):
        if isinstance(results, types.GeneratorType):
            results = tuple(results)

        self._validate_result_type(results, *args)
        self._validate_result_batch_length(results, *args)
        return results


class Aggregator(UDFBase):
    def __init__(self, params=None, output=None, batch=1):
        output = output or {}
        super().__init__(params, DatasetRow.schema | output, batch)

    def validate_results(self, results, *args, **kwargs):
        if isinstance(results, types.GeneratorType):
            for r in results:
                self._validate_result_length(r, *args)
                yield r
            return

        self._validate_result_type(results, *args)
        self._validate_result_batch_length(results, *args)
        return results


class GroupMapper(UDFBase):
    def __init__(self, params=None, output=None, batch=1):
        if batch > 1:
            raise ValueError("GroupMapper does not support batch")
        super().__init__(params, output)

    def validate_results(self, results, *args, **kwargs):
        self._validate_result_type(results, *args)
        self._validate_result_length(results, *args)
        return results



import logging
from argparse import ArgumentParser, ArgumentTypeError

import pytest

from dvcx.cli import (
    TTL_HUMAN,
    TTL_INT,
    find_columns_type,
    get_logging_level,
    get_parser,
    human_time_type,
)
from dvcx.cli_utils import CommaSeparatedArgs, KeyValueArgs


def test_human_time_type():
    assert human_time_type(TTL_HUMAN) == TTL_INT
    assert human_time_type("1h") == 60 * 60
    assert human_time_type("30s") == 30
    assert human_time_type("2w") == 2 * 7 * 24 * 60 * 60

    assert human_time_type("", can_be_none=True) is None

    with pytest.raises(ArgumentTypeError):
        human_time_type("bogus")


def test_find_columns_type():
    assert find_columns_type("") == ["path"]
    assert find_columns_type("du") == ["du"]
    assert find_columns_type("", default_colums_str="name") == ["name"]
    assert find_columns_type("du, name,PATH") == ["du", "name", "path"]

    with pytest.raises(ArgumentTypeError):
        find_columns_type("bogus")


def test_cli_parser():
    parser = get_parser()

    args = parser.parse_args(("ls", "s3://example-bucket/", "--ttl", "1d"))

    assert args.ttl == 24 * 60 * 60
    assert args.sources == ["s3://example-bucket/"]

    assert args.quiet == 0
    assert args.verbose == 0

    assert get_logging_level(args) == logging.INFO

    args = parser.parse_args(("ls", "s3://example-bucket/", "-vvv"))

    assert args.quiet == 0
    assert args.verbose == 3

    assert get_logging_level(args) == logging.DEBUG

    args = parser.parse_args(("ls", "s3://example-bucket/", "-q"))

    assert args.quiet == 1
    assert args.verbose == 0

    assert get_logging_level(args) == logging.CRITICAL


@pytest.mark.parametrize(
    "param,parsed",
    (
        ("p1", ["p1"]),
        ("p1,p2", ["p1", "p2"]),
    ),
)
def test_comma_separated_args(param, parsed):
    parser = ArgumentParser()
    parser.add_argument("--param", default=[], action=CommaSeparatedArgs)

    args = parser.parse_args(("--param", param))
    assert args.param == parsed


@pytest.mark.parametrize("param", (None, ""))
def test_comma_separated_args_error(param):
    parser = ArgumentParser()
    parser.add_argument("--param", default=[], action=CommaSeparatedArgs)

    cmd = ["--param"]
    if param:
        cmd.append(param)
    with pytest.raises(SystemExit):
        parser.parse_args(cmd)


@pytest.mark.parametrize(
    "params,parsed",
    (
        ([], None),
        (["p1=foo"], {"p1": "foo"}),
        (["p1=bar", "p2=baz"], {"p1": "bar", "p2": "baz"}),
        (["p1=foo", "p1=bar"], {"p1": "bar"}),
        (["p1=foo", "p1=bar"], {"p1": "bar"}),
    ),
)
def test_key_value_args(params, parsed):
    parser = ArgumentParser()
    parser.add_argument("--param", nargs=1, action=KeyValueArgs)

    cmd = []
    for p in params:
        cmd.extend(["--param", p])

    args = parser.parse_args(cmd)
    assert args.param == parsed


@pytest.mark.parametrize("param", (None, "p1", "=", "p1=", "=foo"))
def test_key_value_args_error(param):
    parser = ArgumentParser()
    parser.add_argument("--param", nargs=1, action=KeyValueArgs)

    cmd = ["--param"]
    if param:
        cmd.append(param)
    with pytest.raises(SystemExit):
        parser.parse_args(cmd)



import open_clip
import torch
from torch.nn.functional import cosine_similarity
from torch.utils.data import DataLoader

from dvcx.lib.dataset import C, Dataset
from dvcx.lib.feature import ShallowFeature
from dvcx.lib.feature_udf import FeatureMapper
from dvcx.lib.file import File, TextFile
from dvcx.lib.image import ImageReader
from dvcx.lib.text import TextReader

source = "gcs://dvcx-50k-laion-files/000000/00000000*"


class Stem(ShallowFeature):
    stem: str


class StemMapper(FeatureMapper):
    def process(self, data):
        return Stem(stem=data.name.split(".")[0])


class Caption(ShallowFeature):
    caption: str


class CaptionMapper(FeatureMapper):
    def process(self, data):
        return Caption(caption=data.get_value())


def create_dataset():
    imgs = (
        Dataset(source)
        .filter(C.name.glob("*.jpg"))
        .map(StemMapper(inputs=File, outputs=Stem))
    )
    captions = (
        Dataset(source)
        .filter(C.name.glob("*.txt"))
        .map(StemMapper(inputs=File, outputs=Stem))
        .map(CaptionMapper(inputs=TextFile, outputs=Caption))
    )
    return imgs.join(captions.select("stem", "caption"), "stem").save("laion-50k")


if __name__ == "__main__":
    q = create_dataset()

    model, _, preprocess = open_clip.create_model_and_transforms(
        "ViT-B-32", pretrained="laion2b_s34b_b79k"
    )
    tokenizer = open_clip.get_tokenizer("ViT-B-32")

    ds = q.to_pytorch(
        ImageReader(transform=preprocess),
        TextReader("caption", tokenizer=tokenizer),
    )
    loader = DataLoader(ds, batch_size=16)

    similarity_sum = 0
    row_count = 0
    with torch.no_grad(), torch.cuda.amp.autocast():
        for image, text in loader:
            image_features = model.encode_image(image)
            text_features = model.encode_text(text)
            similarity_sum += (
                cosine_similarity(image_features, text_features).sum().item()
            )
            row_count += len(image_features)

    print("Average cosine similarity:", similarity_sum / row_count)



from dvcx.query import C, DatasetQuery
from dvcx.sql.functions import path

source_path = "gcs://dvcx-zalando-hd-resized/zalando-hd-resized/"
ds = (
    DatasetQuery(source_path)
    .filter(C.name.glob("*.jpg"))
    .mutate(**{"class": path.name(C.parent), "label": path.name(path.parent(C.parent))})
)



import os

from openai import OpenAI

from dvcx.query import C, DatasetQuery

query = "What is the best way to do topic modeling?"

k_rag = 5
source_name = "Community"
system_name = "System"

ds = DatasetQuery(name="rag-query").order_by(C.distance).limit(k_rag)

df = ds.to_pandas().sort_values(by="distance", ascending=True)
content = "\n\n".join([f"{source_name}: {page}" for page in df["page"].to_list()])

openai_api_key = os.environ["OPENAI_API_KEY"]
assert openai_api_key.startswith("sk-")
client = OpenAI(api_key=openai_api_key)
del openai_api_key

msg = f"""
{system_name}: {query.strip('?')} according to the {source_name}?

{content}
"""

print(msg)

chat_completion = client.chat.completions.create(
    messages=[
        {
            "role": "user",
            "content": msg.strip(),
        }
    ],
    model="gpt-4",
    temperature=0,
)
response = chat_completion.choices[0].message.content

print("\n\nLLM-RAG Response:\n")
print(response.strip())

ds.limit(k_rag)



import posixpath

import pytest

from dvcx.catalog import Catalog
from dvcx.catalog.catalog import DataSource
from dvcx.node import DirType, Entry
from tests.utils import skip_if_not_sqlite

TREE = {
    "dir1": {
        "d2": {None: ["file1.csv", "file2.csv"]},
        None: ["dataset.csv"],
    },
    "dir2": {None: ["diagram.png"]},
    None: ["users.csv"],
}


def _tree_to_entries(tree: dict, path=""):
    for k, v in tree.items():
        if k:
            dir_path = posixpath.join(path, k)
            yield Entry.from_dir(path, k)
            yield from _tree_to_entries(v, dir_path)
        else:
            for fname in v:
                yield Entry.from_file(path, fname)


@pytest.fixture
def listing(id_generator, metastore, warehouse):
    catalog = Catalog(
        id_generator=id_generator, metastore=metastore, warehouse=warehouse
    )
    lst, _ = catalog.enlist_source("s3://whatever", 1234, skip_indexing=True)
    lst.insert_entry(Entry.root())
    lst.insert_entries(_tree_to_entries(TREE))
    lst.insert_entries_done()
    return lst


def test_resolve_path_in_root(listing):
    node = listing.resolve_path("dir1")
    assert node.dir_type == DirType.DIR
    assert node.is_dir
    assert node.name == "dir1"


def test_path_resolving_nested(listing):
    node = listing.resolve_path("dir1/d2/file2.csv")
    assert node.dir_type == DirType.FILE
    assert node.name == "file2.csv"
    assert not node.is_dir

    node = listing.resolve_path("dir1/d2")
    assert node.dir_type == DirType.DIR
    assert node.is_dir
    assert node.name == "d2"


def test_resolve_not_existing_path(listing):
    with pytest.raises(FileNotFoundError):
        listing.resolve_path("dir1/fake-file-name")


def test_resolve_root(listing):
    node = listing.resolve_path("")
    assert node.dir_type == DirType.DIR
    assert node.is_dir
    assert node.name == ""


def test_path_starts_with_slash(listing):
    node = listing.resolve_path("/dir1")
    assert node.dir_type == DirType.DIR
    assert node.is_dir
    assert node.name == "dir1"


def test_dir_ends_with_slash(listing):
    node = listing.resolve_path("dir1/")
    assert node.dir_type == DirType.DIR
    assert node.is_dir
    assert node.name == "dir1"


def test_file_ends_with_slash(listing):
    with pytest.raises(FileNotFoundError):
        listing.resolve_path("dir1/dataset.csv/")


def _match_filenames(nodes, expected_names):
    assert len(nodes) == len(expected_names)
    names = (node.name for node in nodes)
    assert set(names) == set(expected_names)


def test_basic_expansion(listing):
    nodes = listing.expand_path("*")
    _match_filenames(nodes, ["dir1", "dir2", "users.csv"])


def test_subname_expansion(listing):
    nodes = listing.expand_path("di*/")
    _match_filenames(nodes, ["dir1", "dir2"])


def test_multilevel_expansion(listing):
    skip_if_not_sqlite()
    nodes = listing.expand_path("dir[1,2]/d*")
    _match_filenames(nodes, ["dataset.csv", "diagram.png", "d2"])


def test_expand_root(listing):
    nodes = listing.expand_path("")
    assert len(nodes) == 1
    assert nodes[0].dir_type == DirType.DIR
    assert nodes[0].is_dir


def test_list_dir(listing):
    dir1 = listing.resolve_path("dir1/")
    names = listing.ls_path(dir1, ["name"])
    assert {n[0] for n in names} == {"d2", "dataset.csv"}


def test_list_file(listing):
    file = listing.resolve_path("dir1/dataset.csv")
    src = DataSource(listing, file)
    results = list(src.ls(["id", "name", "dir_type"]))
    assert {r[1] for r in results} == {"dataset.csv"}
    assert results[0][0] == file.id
    assert results[0][1] == file.name
    assert results[0][2] == DirType.FILE


def test_subtree(listing):
    dir1 = listing.resolve_path("dir1/")
    nodes = listing.subtree_files(dir1)
    subtree_files = ["dataset.csv", "file1.csv", "file2.csv"]
    _match_filenames([nwp.n for nwp in nodes], subtree_files)



import json
import os

import fsspec
import mediapipe as mp
import numpy as np
from google.protobuf.json_format import MessageToDict, ParseDict
from mediapipe import solutions
from mediapipe.framework.formats import landmark_pb2
from mediapipe.tasks import python
from mediapipe.tasks.python import vision
from PIL import Image, ImageFile
from tabulate import tabulate

from dvcx.catalog import get_catalog
from dvcx.query import C, DatasetQuery, DatasetRow, Stream, udf
from dvcx.sql.types import JSON


def load_image(stream):
    with stream:
        img = Image.open(stream)
        img.load()
    format = img.format  # typically, this will be JPEG
    image = mp.Image(image_format=mp.ImageFormat.SRGB, data=np.asarray(img))
    return image, format


def landmarks_list_to_pb2(pose_landmarks):
    pose_landmarks_proto = landmark_pb2.NormalizedLandmarkList()
    pose_landmarks_proto.landmark.extend(
        [
            landmark_pb2.NormalizedLandmark(x=landmark.x, y=landmark.y, z=landmark.z)
            for landmark in pose_landmarks
        ]
    )
    return pose_landmarks_proto


def pb2_to_dict(pose_landmarks_proto):
    pose_dict = MessageToDict(pose_landmarks_proto)
    # Test round trip while we are at it
    assert pose_landmarks_proto == ParseDict(
        pose_dict, landmark_pb2.NormalizedLandmarkList()
    )
    return pose_dict


def annotate_image(image, pose):
    annotated_image = np.copy(image)
    for pose_landmarks_proto in pose:
        solutions.drawing_utils.draw_landmarks(
            annotated_image,
            pose_landmarks_proto,
            solutions.pose.POSE_CONNECTIONS,
            solutions.drawing_styles.get_default_pose_landmarks_style(),
        )
    return annotated_image


def mask_image(detection_result):
    # This only takes the most confident detection.
    # We could union all the detections or do a generator over each.
    segmentation_mask = detection_result.segmentation_masks[0].numpy_view()
    visualized_mask = np.repeat(segmentation_mask[:, :, np.newaxis], 3, axis=2) * 255
    return visualized_mask.astype(np.uint8)


@udf(
    params=(Stream(), *tuple(DatasetRow.schema.keys())),
    output={**DatasetRow.schema, "pose": JSON},
)
class PoseDetector:
    annotated_folder = "annotated_images"
    mask_folder = "mask_images"

    def __init__(
        self,
        *,
        model_asset_path,
        bucket_name,
        prefix,
    ):
        base_options = python.BaseOptions(model_asset_path=model_asset_path)
        options = vision.PoseLandmarkerOptions(
            base_options=base_options, output_segmentation_masks=True
        )
        self.detector = vision.PoseLandmarker.create_from_options(options)

        catalog = get_catalog()
        self.client, _ = catalog.parse_url(os.path.join(prefix, bucket_name))

    def save(self, image, source, folder, name, format):
        # Do writeback
        blob_name = os.path.join(folder, name)
        urlpath = os.path.join(source, blob_name)
        cloud_file = fsspec.open(urlpath=urlpath, mode="wb")
        with cloud_file as fp:
            image.save(fp, format=format)

        # Get the blob info
        info_ = self.client.fs.info(urlpath)
        entry = self.client.convert_info(info_, folder)
        return DatasetRow.create(
            name=name,
            source=source,
            parent=folder,
            size=entry.size,
            location=None,
            vtype="",
            dir_type=0,
            owner_name=entry.owner_name,
            owner_id=entry.owner_id,
            is_latest=entry.is_latest,
            last_modified=entry.last_modified,
            version=entry.version,
            etag=entry.etag,
        )

    def __call__(
        self,
        stream,
        *args,
    ):
        # Build a dict from row contents
        record = dict(zip(DatasetRow.schema.keys(), args))

        # Don't re-apply analysis to output
        if record["parent"] in (self.annotated_folder, self.mask_folder):
            return

        # CLeanup to records
        del record["random"]  # random will be populated automatically
        record["is_latest"] = record["is_latest"] > 0  # needs to be a bool
        row = DatasetRow.create(**record)

        # Put into media pipe object
        image, image_format = load_image(stream)

        # Do the detection
        detection_result = self.detector.detect(image)

        # Move into protobuf list for next step
        pose = [landmarks_list_to_pb2(lm) for lm in detection_result.pose_landmarks]

        # Turn into json
        pose_json = [pb2_to_dict(lmp) for lmp in pose]

        # Yield same row back (with json pose info)
        yield (*row, json.dumps(pose_json))

        # No detections ==> we can stop here
        if len(pose) == 0:
            return

        # Annotate image
        annotated_image = annotate_image(image.numpy_view(), pose)
        annotated_image = Image.fromarray(annotated_image)  # make PIL object

        # Save the image and get the cloud object info
        row = self.save(
            image=annotated_image,
            source=record["source"],
            folder=self.annotated_folder,
            name=record["name"],
            format=image_format,
        )
        yield (*row, json.dumps(pose_json))

        # Now do mask
        visualized_mask = mask_image(detection_result)
        visualized_mask = Image.fromarray(visualized_mask)  # make PIL object

        # Save the image and get the cloud object info
        row = self.save(
            image=visualized_mask,
            source=record["source"],
            folder=self.mask_folder,
            name=record["name"],
            format=image_format,
        )
        yield (*row, json.dumps(pose_json))


cloud_prefix = "s3://"  # for GCP just switch to "gcs://"
bucket = "dvcx-50k-laion-files-writable"  # which bucket to use for both read and write
bucket_region = "us-east-2"  # no need to specify for GCP

# Use: !wget -O pose_landmarker.task -q https://storage.googleapis.com/mediapipe-models/pose_landmarker/pose_landmarker_heavy/float16/1/pose_landmarker_heavy.task
model_asset_path = "pose_landmarker.task"

file_type = "*.jpg"  # which files to use
filter_mod = 512  # how much of a subset of the data to use, i.e., 1/512
chunk_num = 1

# only needed for AWS (no effect if using GCP)
os.environ["AWS_REGION"] = bucket_region
os.environ["AWS_DEFAULT_REGION"] = bucket_region

ImageFile.LOAD_TRUNCATED_IMAGES = True

assert chunk_num < filter_mod

pose_udf = PoseDetector(
    model_asset_path=model_asset_path,
    bucket_name=bucket,
    prefix=cloud_prefix,
)

if __name__ == "__main__":
    data = (
        DatasetQuery(os.path.join(cloud_prefix, bucket))
        .filter(C.name.glob(file_type))
        .filter(C.random % filter_mod == chunk_num)
        .generate(pose_udf)
        .results()
    )

    # Output the contents of the new dataset.
    print(tabulate(data))



import os
import os.path
import subprocess
from textwrap import dedent

import pytest

MNT_DATASET_ROWS = dedent(
    """
    01375.png
    07510.png
    08433.png
    mnist-info.txt
    readme.md
    val-info.md
    train-info.md
    02201.png
    02297.png
    08967.png
    01837.png
    05385.png
    09416.png
    01805.png
    06366.png
    08394.png
    00422.png
    04813.png
    05508.png
    07576.png
    08747.png
    00266.png
    06810.png
    08537.png
    09396.png
    09846.png
    00271.png
    00577.png
    01608.png
    07194.png
    08051.png
    """
)

MNT_FILE_TREE = {
    "01375.png": 324,
    "07510.png": 308,
    "08433.png": 224,
    "mnist-info.txt": 0,
    "readme.md": 1193,
    "train": {
        "1": {
            "00266.png": 186,
            "06810.png": 175,
            "08537.png": 206,
            "09396.png": 168,
            "09846.png": 208,
        },
        "2": {
            "00422.png": 257,
            "04813.png": 266,
            "05508.png": 287,
            "07576.png": 297,
            "08747.png": 319,
        },
        "3": {
            "00271.png": 303,
            "00577.png": 293,
            "01608.png": 243,
            "07194.png": 300,
            "08051.png": 304,
        },
        "train-info.md": 18,
    },
    "val": {
        "1": {
            "01837.png": 182,
            "05385.png": 182,
            "09416.png": 162,
        },
        "2": {
            "02201.png": 283,
            "02297.png": 274,
            "08967.png": 333,
        },
        "3": {
            "01805.png": 345,
            "06366.png": 302,
            "08394.png": 329,
        },
        "val-info.md": 20,
    },
}

# Note that these commands are tested in order.
E2E_STEPS = (
    {
        "command": (
            "dvcx",
            "ls",
            "--anon",
            "s3://ldb-public/remote/datasets/dogs-and-cats/",
        ),
        "expected": dedent(
            """
            dogs-and-cats.tar.gz
            dogs-and-cats.zip
            license
            """
        ),
        "listing": True,
    },
    {
        "command": (
            "dvcx",
            "du",
            "--anon",
            "s3://ldb-public/remote/datasets/dogs-and-cats/",
        ),
        "expected": "   9.2M s3://ldb-public/remote/datasets/dogs-and-cats/\n",
    },
    {
        "command": (
            "dvcx",
            "find",
            "--iname",
            "*DOG*",
            "--anon",
            "s3://ldb-public/remote/datasets/",
        ),
        "expected": dedent(
            """
            s3://ldb-public/remote/datasets/Stanford-dog-breeds/
            s3://ldb-public/remote/datasets/dogs-and-cats/
            s3://ldb-public/remote/datasets/dogs-and-cats/dogs-and-cats.tar.gz
            s3://ldb-public/remote/datasets/dogs-and-cats/dogs-and-cats.zip
            """
        ),
        "listing": True,
    },
    {
        "command": ("dvcx", "du", "--anon", "s3://ldb-public/remote/datasets/"),
        "expected": "  43.3G s3://ldb-public/remote/datasets/\n",
    },
    {
        "command": (
            "dvcx",
            "cp",
            "-r",
            "--anon",
            "s3://ldb-public/remote/datasets/mnist-tiny/",
            "mnt-cp",
        ),
        "expected": "",
        "downloading": True,
        "instantiating": True,
        "files": {
            "mnt-cp": MNT_FILE_TREE,
        },
    },
    {
        "command": (
            "dvcx",
            "clone",
            "-r",
            "--anon",
            "s3://ldb-public/remote/datasets/mnist-tiny/",
            "mnt",
        ),
        "expected": "",
        "downloading": True,
        "instantiating": True,
        "files": {
            "mnt": MNT_FILE_TREE,
        },
    },
    {
        "command": ("dvcx", "ls-datasets"),
        "expected": "mnt (v1)\n",
    },
    {
        "command": ("dvcx", "ls-dataset-rows", "mnt", "--version", "1"),
        "expected": MNT_DATASET_ROWS,
        "sort_expected_lines": True,
    },
    {
        "command": ("dvcx", "ls-datasets"),
        "expected": "mnt (v1)\n",
    },
    {
        "command": ("dvcx", "ls-dataset-rows", "mnt", "--version", "1"),
        "expected": MNT_DATASET_ROWS,
        "sort_expected_lines": True,
    },
    {
        "command": ("dvcx", "edit-dataset", "mnt", "--new-name", "mnt-new"),
        "expected": "",
    },
    {
        "command": ("dvcx", "ls-datasets"),
        "expected": "mnt-new (v1)\n",
    },
    {
        "command": ("dvcx", "ls-dataset-rows", "mnt-new", "--version", "1"),
        "expected": MNT_DATASET_ROWS,
        "sort_expected_lines": True,
    },
    {
        "command": ("dvcx", "rm-dataset", "mnt-new", "--version", "1"),
        "expected": "",
    },
    {
        "command": ("dvcx", "ls-datasets"),
        "expected": "",
    },
    {
        "command": ("dvcx", "gc"),
        "expected": "Nothing to clean up.\n",
    },
)


def verify_files(files, base=""):
    """Recursively validate file and directory structure."""
    for name, value in files.items():
        full_name = os.path.join(base, name)
        if isinstance(value, dict):
            assert os.path.isdir(full_name)
            verify_files(value, full_name)
        else:
            assert os.path.isfile(full_name)
            assert os.path.getsize(full_name) == value


def run_step(step):
    """Run an end-to-end test step with a command and expected output."""
    result = subprocess.run(
        step["command"],
        shell=False,  # noqa: S603
        capture_output=True,
        check=True,
        encoding="utf-8",
    )
    if step.get("sort_expected_lines"):
        assert sorted(result.stdout.split("\n")) == sorted(
            step["expected"].lstrip("\n").split("\n")
        )
    else:
        assert result.stdout == step["expected"].lstrip("\n")
    if step.get("listing"):
        assert "Listing" in result.stderr
    else:
        assert "Listing" not in result.stderr
    if step.get("downloading"):
        assert "Downloading" in result.stderr
    else:
        assert "Downloading" not in result.stderr
    if step.get("instantiating"):
        assert "Instantiating" in result.stderr
    else:
        assert "Instantiating" not in result.stderr
    files = step.get("files")
    if files:
        verify_files(files)


@pytest.mark.e2e
def test_cli_e2e(tmp_dir):
    """End-to-end CLI Test"""
    for step in E2E_STEPS:
        run_step(step)



from dvcx.lib.feature import Feature
from dvcx.lib.file import TextFile
from dvcx.lib.reader import FeatureReader, LabelReader


class NameFeature(Feature):
    name: str

    def get_value(self):
        return self.name


class PrefixReader(FeatureReader):
    def __call__(self, value):
        return "prefix-" + value


class SuffixReader(FeatureReader):
    def __init__(self, fr_class, suffix):
        self.suffix = suffix
        super().__init__(fr_class)

    def __call__(self, value):
        return value + self.suffix


def test_feature_reader():
    reader = PrefixReader(NameFeature)
    feature = NameFeature(name="name")
    assert reader(feature.get_value()) == "prefix-name"


def test_feature_reader_args():
    reader = SuffixReader(NameFeature, suffix="-suffix")
    feature = NameFeature(name="name")
    assert reader(feature.get_value()) == "name-suffix"


def test_feature_reader_file(tmp_path, catalog):
    file_name = "myfile"
    data = "myText"

    file_path = tmp_path / file_name
    with open(file_path, "w") as fd:
        fd.write(data)

    reader = PrefixReader(TextFile)
    file = TextFile(name=file_name, source=f"file://{tmp_path}")
    file.set_catalog(catalog)

    with open(file_path) as fd:
        file.set_file(fd, caching_enabled=False)
        assert reader(file.get_value()) == "prefix-myText"


def test_feature_reader_column():
    reader = PrefixReader("name")
    feature = NameFeature(name="name")
    assert reader(feature.get_value()) == "prefix-name"


def test_label_reader():
    classes = ["my", "name"]
    reader = LabelReader("name", classes=classes)
    feature = NameFeature(name="name")
    assert reader(feature.get_value()) == 1



"""
This is a simple UDF to demonstrate local parallel processing with multiprocessing.

In add_signals specify either parallel=-1 to use processes equal to the number
of CPUs/cores on your current machine, or parallel=N for N processes.
The default if parallel is not specified is to run single-threaded.

The UDF specified in add_signals will then be run in parallel across all these
worker processes, no other code changes are needed.

Benchmark results showed an almost 8X speedup on a MacBook Pro, using this test, with
parallel processing reducing execution time from a median of 377s to 48s total.

To install script dependencies: pip install tabulate
"""

from sqlalchemy import Integer
from tabulate import tabulate

from dvcx.query import C, DatasetQuery, udf


# This is a simple single-threaded benchmark function to demonstrate the speedup
# that can be achieved with multiprocessing, by enabling parallel in add_signals.
def fibonacci(n):
    if n <= 1:
        return n
    return fibonacci(n - 1) + fibonacci(n - 2)


# Define the UDF:
@udf(
    ("name",),  # Columns consumed by the UDF.
    {
        "path_len": Integer
    },  # Signals being returned by the UDF, with the signal name and type.
)
def name_len_benchmark(name):
    # Run the fibonacci benchmark as an example of a single-threaded CPU-bound UDF
    fibonacci(35)
    if name.endswith(".json"):
        return (-1,)
    return (len(name),)


# Save as a new dataset
DatasetQuery(
    path="gcs://dvcx-datalakes/dogs-and-cats/",
    anon=True,
).filter(C.name.glob("*cat*")).add_signals(name_len_benchmark, parallel=-1).save(
    "cats_with_signal"
)

# Output the contents of the new dataset.
print(tabulate(DatasetQuery(name="cats_with_signal").results()[:10]))



from queue import Empty, Full
from typing import Optional

from dvcx.query.dispatch import (
    STOP_SIGNAL,
    UDFDispatcher,
    get_from_queue,
    put_into_queue,
)


class MockQueue:
    def __init__(self) -> None:
        self.return_empty_once = True
        self.return_full_once = True
        self.put_signal: Optional[str] = None
        self.put_count: int = 0

    def put_nowait(self, task: str) -> None:
        if self.return_full_once:
            self.return_full_once = False
            raise Full
        assert task == STOP_SIGNAL
        self.put_signal = task
        self.put_count += 1

    def get_nowait(self) -> str:
        if self.return_empty_once:
            self.return_empty_once = False
            raise Empty
        return STOP_SIGNAL


def test_get_from_queue():
    mock_queue = MockQueue()

    assert get_from_queue(mock_queue) == STOP_SIGNAL


def test_put_into_queue():
    mock_queue = MockQueue()

    assert put_into_queue(mock_queue, STOP_SIGNAL) is None
    assert mock_queue.put_signal == STOP_SIGNAL


def test_send_stop_signal_to_workers():
    mock_queue = MockQueue()

    UDFDispatcher.send_stop_signal_to_workers(mock_queue, 8)

    assert mock_queue.put_signal == STOP_SIGNAL
    assert mock_queue.put_count == 8
